/***************************************************************************
*  PROGRAM: qidx.p
*   AUTHOR: Stephen Hawkins
*     DATE: 26 March 2006
* INCLUDES: qidx-tt.i 
*     RUNS: qidx-db.p
*   RUN BY: Can be run directly from Progress editor or any other program.
*  PURPOSE: Allow user to specify a table definition and a query on this,
*           then report detailed index information for this query. The aim
*           of this program is to accurately and comprehensively explain
*           compiler index selection.
* CONTENTS: - &Preprocessing
*           - Define shared temp-tables
*           - Define temp-tables:
*             - t-where-tree
*             - t-test-index
*             - t-test-fem-set
*             - t-index-tree
*           - Defined program variables
*           - Defined variables and form for frame f
*           - UI triggers:
*             - ON entry of fields or go of frame f
*             - ON leave of fields
*             - ON help of frame f
*           - Run main
*           - Defined Functions:
*             - FUNCTION next-wtree-child
*             - FUNCTION next-itree-child
*             - FUNCTION valid-field
*             - FUNCTION clean-whitespace
*             - FUNCTION combination
*           - Internal procedures:
*             - PROCEDURE main
*             - PROCEDURE load-schema
*             - PROCEDURE validate-v6-query
*             - PROCEDURE validate-db-name
*             - PROCEDURE validate-file-name
*             - PROCEDURE validate-def-table
*             - PROCEDURE validate-buffer-name
*             - PROCEDURE validate-where-expr
*             - PROCEDURE validate-index-use
*             - PROCEDURE validate-by-fields
*             - PROCEDURE parse-table-def
*             - PROCEDURE parse-table-like
*             - PROCEDURE parse-table-tidy
*             - PROCEDURE parse-where-expr
*             - PROCEDURE parse-where-bracket
*             - PROCEDURE parse-where-or
*             - PROCEDURE parse-where-single
*             - PROCEDURE parse-where-escaped
*             - PROCEDURE parse-where-e-state
*             - PROCEDURE analyse-where-tree
*             - PROCEDURE analyse-where-index
*             - PROCEDURE analyse-where-undo
*             - PROCEDURE analyse-indexes
*             - PROCEDURE analyse-fem-sets
*             - PROCEDURE select-indexes
*             - PROCEDURE tidy-post-analysis
*             - PROCEDURE tidy-where-tree
*             - PROCEDURE report-qidx-info
*             - PROCEDURE report-where-tree
*             - PROCEDURE report-index-tree
*             - PROCEDURE report-index-analysis
*             - PROCEDURE report-table-schema
***************************************************************************/

/***************************************************************************
*                           &PREPROCESSING
* 
* &tt-index-order: 
*   Traditionally, the last rule of index selection is: "Select the first
*    index alphabetically by index name" but this is not strictly correct.
*   More correct is: "Select the first index by index order".
*   In the case of database tables, index order is always the same as
*    alphabetic order of index names.
*   In the case of temp-tables, index order may be alphabetic order or
*    define order (the order indexes are defined in temp-table definition)
*    depending on the version of Progress. In V7.3 and V8.0 it seems to be
*    alphabetic order. In V9.1 it is define order. I don't know about other
*    versions. I'm guessing the order method changed somewhere between V8.0
*    and V9.1 - perhaps V9.0?
*   Change this if necessary.
***************************************************************************/

&if integer( entry( 1, proversion, "." ) ) >= 9 &then
  &global-define tt-index-order define
&else
  &global-define tt-index-order alpha
&endif


/***************************************************************************
*                      DEFINE SHARED TEMP-TABLES
*
* qidx-tt.i contains temp-table defintions for t-table, t-field, t-index,
* and t-ixfld.
* These temp-tables are shared with qidx-db.p which has to be run as a
* separate program.
***************************************************************************/

{qidx-tt.i "new shared"}  


/***************************************************************************
*                    DEFINE TEMP-TABLE T-WHERE-TREE
* 
* Collectively, these records represent the where-tree analysis of the
*  WHERE expression - each record represents a node in the where-tree.
* - wid: a unique identifier for each where-tree node
* - parent-wid: identifies parent node (zero if no parent)
* - child-num: child order of parent
* - inode-wid:
* - type: "AND", "OR", or "LEAF"
* - expr: expression text for a leaf node
* - field-name: name of indexable field in expression if e-state begins "I"
* - e-state: state of expression based on operator - possible values are:
*   - IEQ: Indexable Equality (=,EQ)
*   - IRB: Indexable Range-Both (BEGINS)
*   - IRL: Indexable Range-Left (>=,>,GE,GT)
*   - IRR: Indexable Range-Right (<=,<,LE,LT)
*   - IW:  Indexable Word-Index (CONTAINS)
*   - IID: (pseudo-)Indexable ID (RECID,ROWID)
*   - X:   expression is excluded from index analysis
* - i-state: index state - possible values are:
*   - "I": Indexable node (inode) with index analysis
*   - "<": leaf child of AND-inode
*   - "+": OR-node included in index-tree without being an inode itself.
*   - "-": non-leaf node not included in index-tree, but children may be.
*   - "X": node is excluded from index analysis along with all its children.
* - comment: can help further describe e-state or i-state.
* - word-index: name of word-index or "*" if node has word-index descendant.
* - whole-wid: = wid if whole-index inode, or wid of whole-index child if
*              OR-node (zero if not whole-index)
* - yes-replace: is expression replaced by YES in post-index filter?
***************************************************************************/

def temp-table t-where-tree
  field wid as int
  field parent-wid as int
  field child-num as int
  field inode-wid as int
  field type as char
  field expr as char format "x(20)"
  field field-name as char
  field e-state as char
  field i-state as char
  field comment as char
  field word-index as char
  field whole-wid as int
  field yes-replace as log
  index wid is unique wid
  index parent+child is unique parent-wid child-num
  index inode+field+estate inode-wid field-name e-state.


/***************************************************************************
*                    DEFINE TEMP-TABLE T-TEST-INDEX
* 
* Contains test analysis data for each non-word index, for each inode of 
*  the where-tree
* - inode-wid: link to indexable node of where-tree
* - index-name: name of non-word index or "RECID" or "{FEM-set}"
* - index-order: order of index of table
* - is-primary: is this the primary index of table?
* - num-equality: number of equality matches
* - num-range: number of range matches
* - num-sort: number of sort matches
* - fem: Full Equality Match type
*   - "N" non-unique FEM index
*   - "U" unique FEM index
*   - "W" word-index
*   - "R" RECID pseudo-index
*   - "S" FEM-set pseudo-index entity
*   - "" non-FEM index
* - fem-set: comma separated set of indexes in set (if fem = "S")
* - is-selected: was this test-index selected for this inode?
***************************************************************************/

def temp-table t-test-index
  field inode-wid as int
  field index-name as char
  field index-order as int
  field index-type as char
  field is-primary as log
  field num-equality as int
  field num-range as int
  field num-sort as int
  field fem as char
  field fem-set as char
  field is-selected as log
  index wid+name is unique inode-wid index-name.


/***************************************************************************
*                    DEFINE TEMP-TABLE T-TEST-FEM-SET
* 
* If an inode has more than one non-unique FEM index then a test-fem-set is
*  created for each combination of non-unique FEM-indexes. The best
*  test-fem-set will be selected to create a FEM-set test-index.
* - inode-wid: link to indexable node of where-tree
* - index-list: list of indexes in this particular combination
* - field-list: list of distinct fields covered by index-list
* - sum-fields: the sum of fields in each index in index-list
* - combi-num: combination order (based on index-order)
* - key: for additional  fields (currently only used for "max" for the set
*        that has the complete set of num-unique fem indexes)
***************************************************************************/

def temp-table t-test-fem-set
  field inode-wid as int
  field index-list as char
  field field-list as char
  field sum-fields as int
  field combi-num as int
  field key as char
  index inode-wid inode-wid
  index key is word-index key.


/***************************************************************************
*                    DEFINE TEMP-TABLE T-INDEX-TREE
* 
* Collectively, these records represent the index-tree of indexes selected
*  by the compiler. It has some relation to the where-tree. It's structure
*  may be similar to the where-tree or it could be quite different,
*  depending
* - iid: unique identifier for index-tree node.
* - parent-iid: identifies parent node (zero if no parent)
* - child-num: child order of parent
* - type: one of (AND,OR,LEAF)
* - orig-wid: link to original where-tree node (zero if leaf child of AND)
* - inode-wid: link to inode of where-tree this node belongs to
* - word-wid: link to CONTAINS expression in where-tree if word-index, or
*             -1 if node has word-index descendant (zero if not word-index)
* - whole-wid: = inode-wid if whole-index, or inode-wid of whole-index child
*              if OR-node (zero if not whole-index)
* - is-fem: does this node have FEM flag set in compiled index-tree?
* - index-name: name of index or RECID if leaf node.
* - set-flag: 3 bits indicate membership of FEM-set
*   - bit 1: member of FEM-set?
*   - bit 2: first member of FEM-set?
*   - bit 3: last member of FEM-set?
* - reason: description why the index of this node was selected.
***************************************************************************/

def temp-table t-index-tree
  field iid as int
  field parent-iid as int
  field child-num as int
  field type as char
  field inode-wid as int
  field orig-wid as int
  field word-wid as int
  field whole-wid as int
  field is-fem as log
  field index-name as char
  field set-flag as int
  field reason as char
  index iid is unique iid
  index parent+child is unique parent-iid child-num
  index inode inode-wid.


/***************************************************************************
*                       DEFINE PROGRAM VARIABLES
* 
* - wid is a program sequence for t-where-tree.wid.
* - root-wid is the wid of the root node of the where-tree.
* - iid is a program sequence for t-index-tree.iid.
* - root-iid is the iid of the root node of the index-tree
* - table-name is the full name of the current table.
* - index-by-alpha = yes if index-order of current table is alphabetic.
***************************************************************************/
  
def var wid as int.
def var root-wid as int.
def var iid as int.
def var root-iid as int.
def var table-name as char.
def var index-by-alpha as log.


/***************************************************************************
*                     DEFINE VARS AND FORM FOR FRAME F
* 
***************************************************************************/

def var v6-query as log.
def var db-name as char.
def var file-name as char.
def var def-table as char.
def var index-use as char.
def var buffer-name as char.
def var where-expr as char.
def var by-fields as char.

form
  db-name label "DB" colon 7 format "x(26)"
    help "LDBName or Alias of Database (or blank if Temp-table)"
  file-name label "Table" format "x(32)"
    help "Enter name of table to query"
  skip
  def-table label "Define" colon 7 view-as editor size 70 by 10
    help "Table definition. Press Help key for more info."
  buffer-name label "Buffer" colon 7 format "x(65)"
    help "Buffer name"
  where-expr label "Where" colon 7 view-as editor size 70 by 5
    help "Enter simplified WHERE expr. Press HELP key for more info."
  skip
  by-fields label "By" colon 7 view-as editor size 70 by 2
    help "Enter list of BY fields. Press HELP key for more info."
  skip
  v6-query label "V6" colon 7
    help "Yes if V6 query or FIND statement; otherwise no"
  index-use label "Use-Index" format "x(32)" 
    help "Enter index name if USE-INDEX option used"
  with frame f 
       side-labels 
       three-d.

 
/***************************************************************************
*                        TRIGGER ON ENTRY OR GO
*
* Do not allow go or entry into table sensitive fields until a valid table
* name has been entered.
***************************************************************************/

on entry of def-table, buffer-name, index-use, where-expr, by-fields or
   go of frame f
do:
  if not can-find( t-table where t-table.table-name = table-name ) then
  do:
    apply "entry" to file-name.
    return no-apply.
  end.
  if last-event:function = "go" and
     not can-find( first t-field where t-field.table-name = table-name )
  then do:
    apply "entry" to def-table.
    return no-apply.
  end.
end.


/***************************************************************************
*                          TRIGGER ON LEAVE
* 
* Run a custom validate procedure for each individual field.
* If there is a return-value then alert message this
***************************************************************************/

on leave of v6-query, db-name, file-name, def-table, index-use, 
            buffer-name, where-expr, by-fields
do:
  run value( "validate-" + self:name)( input self:screen-value ).
  if return-value > "" then
    message return-value view-as alert-box.
  if return-value <> "" then
    return no-apply.
end.


/***************************************************************************
*                      TRIGGER ON HELP OF DEF-TABLE
***************************************************************************/

on help of frame f
  message
    "- Use 'DB' and 'Table' fields to specify an existing table or a new"
    "table to define. Leaving 'DB' blank will define a temp-table,"
    "otherwise a database table is assumed."
    skip
    "- The definition of any table can be temporarily created or modified"
    "by editing the 'Define' field, using simplified temp-table syntax:" 
    skip
    "    [LIKE <table-name>]" skip
    "    ~{FIELD <field-name> [AS <datatype>] [EXTENT <number>]} ..." skip
    "    [INDEX <index-name> [IS [PRIMARY|UNIQUE|WORD-INDEX] ...] ..." skip
    "      ~{<field-name> [ASCENDING|DESCENDING]} ... ] ..." skip
    "An interpretation of your definition will be displayed when you leave"
    "'Define' (any unexpected syntax is ignored without error)."
    "The original definition of an existing table you have experimented"
    "with can be retrieved by blanking the 'Define' field."
    skip
    "- The 'Buffer' field allows you to specify a buffer name for the table"
    "in case you want to copy-and-paste a WHERE expression with qualified"
    "field names." skip
    "- The WHERE expression is entered in the 'Where' field. Most"
    "copy-and-pasted preprocessing-free expressions should be"
    "interpreted okay but there are some limitations ..."
    "For indexable phrases to be interpreted correctly, operators"
    "(=,EQ,BEGINS,>,GT,>=,GE,<,LT,<=,LE,CONTAINS) and conjunctions"
    "(AND,OR) must be whitespace separated, and field-names and"
    "RECID/ROWID functions must be on the LHS of the operator."
    "Expressions on the RHS of the operator can be almost anything you"
    "like for reporting purposes - they are not interpreted except to look"
    "for a field reference that will exclude it from indexing -"
    'but be careful if it includes any of these elements: "(,),AND,OR".'
    "If in doubt, replace"
    "the RHS expression with a single descriptive (non-field) name."
    "Any phrase not accepted as indexable is assumed to be a valid"
    "non-indexable phrase."
    skip
    "- If your query has sorting then enter BY fields in the 'By' field."
    "The first BY phrase that is not a simple field of your query buffer"
    "will be ignored along with any following BY phrases as they do not"
    "affect index selection."
    skip
    "- Specify 'V6' = yes if your query is for version 6, -v6q startup"
    "parameter is used, your query is in a FIND statement, or the USE-INDEX"
    "option is used."
    skip
    "- The 'Use-index' field is provided for completeness. It may be"
    "useful to see if specifying this makes any difference."
    view-as alert-box information.


/***************************************************************************
*                               RUN MAIN
***************************************************************************/
    
run main.

    
/***************************************************************************
*                       FUNCTION NEXT-WTREE-CHILD
* 
* Returns the next available child-num in the where-tree for p-parent-wid.
***************************************************************************/

function next-wtree-child returns int( p-parent-wid as int ):

  def var child-num as int.
  
  def buffer t-wtree for t-where-tree.
  
  find last t-wtree where t-wtree.parent-wid = p-parent-wid no-error.
  child-num = 
    (if available t-wtree then t-wtree.child-num + 1 else 
     if p-parent-wid = ? then ? else 1).
  return child-num.
  
end function.


/***************************************************************************
*                       FUNCTION NEXT-ITREE-CHILD
* 
* Returns the next available child-num in the index-tree for p-parent-iid.
***************************************************************************/

function next-itree-child returns int( p-parent-iid as int ):

  def var child-num as int.
  
  def buffer t-itree for t-index-tree.
  
  find last t-itree where t-itree.parent-iid = p-parent-iid no-error.
  child-num = 
    (if available t-itree then t-itree.child-num + 1 else 
     if p-parent-iid = ? then ? else 1).
  return child-num.
  
end function.


/***************************************************************************
*                         FUNCTION VALID-FIELD
* 
* Check to see if p-name is a valid field-name of table-name.
* If p-name is qualified with a buffer name then the qualification must
*  agree with buffer-name. Otherwise it is assumed to belong to the query
*  buffer and just check that the table has a field of this name.
* Can just check the qualification if the last component of p-name = "*"
*  (e.g. validate RECID( buffer )).
***************************************************************************/

function valid-field returns log ( p-name as char ):

  def var field-name as char.
  def var qualify-name as char.
  
  /* CHECK QUALIFICATION OF P-NAME DOESN'T DISAGREE WITH BUFFER-NAME */
  if index( p-name, "." ) > 0 then
  do:
    assign
      field-name = entry( num-entries( p-name, "." ), p-name, "." )
      qualify-name = 
        substring( p-name, 1, length( p-name ) - length( field-name ) - 1 )
      p-name = field-name.
    if qualify-name <> buffer-name and 
       qualify-name <> entry( 2, buffer-name + ".", "." )
    then 
      return no.
  end.
  
  /* CONFIRM P-NAME IS A VALID FIELD OF CURRENT TABLE OR BLANK */
  return p-name = "*" or
         can-find( t-field where 
           t-field.table-name = table-name and
           t-field.field-name = p-name ).
                    
end function.


/***************************************************************************
*                      FUNCTION CLEAN-WHITESPACE
*
* Converts all sequences of whitespace (space, tab, LF, CF) in p-text to a
*  single space to make parsing easier.
***************************************************************************/

function clean-whitespace returns char( p-text as char ):

  def var c as char.
  
  assign
    c = 
      replace( replace( replace( p-text, "~n", " " ), 
        "~r", " " ), "~t", " " ).
  do while index( c, "  " ) > 0:
    c = replace( c, "  ", " " ).
  end.
  c = trim( c ).
  return c.
  
end function.


/***************************************************************************
*                         FUNCTION COMBINATION
* 
* Returns the "combination" set of p-list taken p-num elements at a time.
* This is designed to be used for index names but it could be used for any
*  list of strings that don't have space or comma characters.
* The pos array limits p-list to a maximum of 20 elements but the 32k record
*  limit of Progress means this function will fail with fewer elements than
*  this (13C7 = 1716 is just big enough for single character elements - if
*  larger than this then results will need to be returned via another way).
***************************************************************************/

function combination returns char( p-list as char, p-num as int ):

  def var pos as int extent 20.
  def var max-pos as int.
  def var ptr as int init 1.
  def var i as int.
  def var list as char.
  def var combi-list as char.
  
  assign
    max-pos = num-entries( p-list )
    list = fill( ",", p-num - 1 ).
  
  /* FIND COMBINATIONS WHILE LEFTMOST POINTER IS STILL IN PLAY */
  do while ptr > 0:

    /* INCREMENT CURRENT POSITION WHILE CAN STILL GO FURTHER RIGHT */
    do while pos[ ptr ] < max-pos - p-num + ptr:
      pos[ ptr ] = pos[ ptr ] + 1.

      /* ANY POSITIONS TO RIGHT OF CURRENT POSITION ARE SHIFTED LEFT,
         THEN CHANGE CURRENT POINTER TO RIGHTMOST */
      if ptr < p-num then
      do:
        do i = ptr + 1 to p-num:
          pos[ i ] = pos[ ptr ] + i - ptr.
        end.
        ptr = p-num.
      end.
      
      /* CONSTRUCT SINGLE LIST AND ADD IT TO COMBI-LIST */
      do i = 1 to p-num:
        entry( i, list ) = entry( pos[ i ], p-list ).
      end.
      combi-list = trim( combi-list + " " + list ).
      
    end.
    
    /* SHIFT POINTER LEFT WHEN CURRENT POSITION IS AT RIGHT END */
    ptr = ptr - 1.
 
  end.
        
  return combi-list.  
   
end function.


/***************************************************************************
*                            PROCEDURE MAIN              
* 
* Table definitions are initially copied from all connected databases to
*  temp-tables.
* Then repeatedly: allow user to specify a table definition and query, parse
*  and analyse the where expression, and report query index information.
***************************************************************************/

procedure main:
  
  /* COPY TABLE DEFINITIONS FROM ALL CONNECTED DATABASES TO TEMP-TABLES */
  run load-schema.
  
  assign
    db-name = ldbname( "dictdb" )
    db-name = (if db-name = ? then "" else db-name)
    index-use:sensitive in frame f = no.
  
  
  repeat: 

    /* INITIALISE */
    empty temp-table t-where-tree.
    empty temp-table t-test-index.
    empty temp-table t-test-fem-set.
    empty temp-table t-index-tree.
    assign  
      wid = 0
      iid = 0
      root-iid = 1.
  
    /* ALLOW USER TO UPDATE TABLE DEFINITION AND QUERY FIELDS */
    update 
      db-name
      file-name
      def-table
      buffer-name
      where-expr
      by-fields
      v6-query
      index-use
      with frame f.

    /* PARSE AND ANALYSE WHERE EXPRESSION */
    run parse-where-expr( where-expr ).
    run analyse-where-tree( root-wid, "", 0 ).
    run tidy-post-analysis.

    /* REPORT QUERY INDEX INFORMATION DETAILS */
    run report-qidx-info.
  
  end. /* repeat */
  
end procedure.
  

/***************************************************************************
*                       PROCEDURE LOAD-SCHEMA
* 
* Create schema data for each table in each connected database.
***************************************************************************/

procedure load-schema:

  def var i as int.
  def var save-dictdb as char.
  def var db-name as char.
  
  save-dictdb = ldbname( "dictdb" ).
 
  do i = 1 to num-dbs:
    db-name = ldbname( i ).
    create alias dictdb for database value( db-name ).
    run qidx-db.p ( db-name, "*" ).
  end.  
  
  if save-dictdb <> ? then  
    create alias dictdb for database value( save-dictdb ).

end procedure.

 
/***************************************************************************
*                      PROCEDURE VALIDATE-DB-NAME
* 
* Give warning message if entered db-name is not a connected database or
*  blank (for temp-table), but allow this anyway in case want to define a
*  database table without any database being connected.
* Initialise all dependent fields if this value changes.
***************************************************************************/

procedure validate-db-name:
  def input param p-name as char.

  def var ok as log.
  
  if p-name = db-name then
    return.
  
  if ldbname( p-name ) = ? and p-name <> "" and
    not can-find( first t-table where t-table.db-name = p-name )
  then do:
    message "Database " + p-name + " is not connected. Use this anyway?"
            view-as alert-box warning buttons yes-no update ok.
    if not ok then
      return ?.
  end.

  assign
    table-name = ""
    file-name = ""
    file-name:screen-value in frame f = ""
    def-table = ""
    def-table:screen-value = ""
    buffer-name = ""
    buffer-name:screen-value = ""
    where-expr = ""
    where-expr:screen-value = ""
    index-use = ""
    index-use:screen-value = ""
    by-fields = ""
    by-fields:screen-value = ""    
    db-name = p-name.
 
  return.
  
end procedure.


/***************************************************************************
*                     PROCEDURE VALIDATE-FILE-NAME
* 
* Make sure table name isn't blank.
* Check if file-name is already a recognised table of db-name. If not then
*  ask user to confirm creation of a new table definition. Otherwise display
*  existing definition and initialise other table dependent fields.
***************************************************************************/


procedure validate-file-name:
  def input param p-name as char.
    
  def var ok as log init yes.
  
  /* DON'T ALLOW BLANK TABLE NAME */
  if p-name = "" then
    return 
      (if last-event:widget-enter <> ? and 
       last-event:widget-enter:name = "db-name" then "" 
       else "You must enter a table name.").
  
  assign
    table-name = 
      (if ldbname( db-name ) = ? then db-name else ldbname( db-name )) + 
      (if db-name = "" then "" else ".") + p-name.
    index-by-alpha = 
      db-name <> "" or "{&tt-index-order}" = "alpha".
  
  /* ALLOW CREATION OF NEW TABLE IF NOT ALREADY DEFINED */
  if not can-find( t-table where t-table.table-name = table-name ) then
  do:
    message 
      "Define new " + (if db-name = "" then "TEMP-" else "DATABASE ") +
      "TABLE " + trim( table-name, "." ) + "?"
      view-as alert-box buttons yes-no
      update ok.
    if not ok then
      return ?.
    create t-table.
    assign
      t-table.db-name = db-name
      t-table.file-name = file-name
      t-table.table-name = table-name
      t-table.from-db = no
      t-table.modified = no.
  end.

  else
  if p-name = file-name then
    return.
  
  /* INITIALISE DEPENDENT FIELDS */  
  run report-table-schema( output def-table ).    
  assign
    file-name = p-name
    def-table:screen-value in frame f = def-table
    buffer-name = table-name
    buffer-name:screen-value = buffer-name
    where-expr = ""
    where-expr:screen-value = ""
    index-use = ""
    index-use:screen-value = ""
    by-fields = ""
    by-fields:screen-value = "".

  return.
  
end procedure.


/***************************************************************************
*                     PROCEDURE VALIDATE-DEF-TABLE
* 
* Parse entered table definition and redisplay interpretation of this.
***************************************************************************/

procedure validate-def-table:
  def input param p-text as char.

  if def-table = p-text then
    return.

  run parse-table-def( clean-whitespace( p-text ) ).
  
  run report-table-schema( output def-table ).
  def-table:screen-value in frame f = def-table.
    
  return.
  
end procedure.


/***************************************************************************
*                    PROCEDURE VALIDATE-BUFFER-NAME
* 
* Make sure that entered buffer-name is only a single name, or an alias of
*  db-name + "." + file-name.
***************************************************************************/

procedure validate-buffer-name:
  def input param p-name as char.

  if buffer-name = p-name then
    return.
  
  if num-entries( p-name, "." ) > 2 then
    return "No more than one dot is allowed in buffer name".
    
  if index( p-name, "." ) > 0 and 
     (ldbname( entry( 1, p-name, "." ) ) <> db-name or 
      entry( 2, p-name, "." ) <> file-name)
  then
    return 
      "Buffer must be a single name or <alias of DB " + db-name + ">." +
      file-name.
  
  buffer-name = p-name.
  return.
  
end procedure.


/***************************************************************************
*                    PROCEDURE VALIDATE-WHERE-EXPR
* 
* Just check that each bracket has a proper match here. Actual parsing of
*  where-expr is delayed until finish updating frame. 
***************************************************************************/

procedure validate-where-expr:
  def input param p-text as char.
  
  def var r-pos as int.
  def var l-pos as int.
  def var q-text as char.
  
  if where-expr = p-text then
    return.
    
  if num-entries( p-text, "(" ) <> num-entries( p-text, ")" ) then
    return "Number of left and right brackets don't match".
  
  q-text = clean-whitespace( p-text ).
  
  do while index( q-text, ")" ) > 0:
    assign
      r-pos = index( q-text, ")" )
      l-pos = r-index( substring( q-text, 1, r-pos ), "(" ).
    if l-pos = 0 then
      return "No matching left bracket found".
    assign
      substring( q-text, r-pos, 1 ) = ""
      substring( q-text, l-pos, 1 ) = "".
  end.
  
  where-expr = p-text.
  return.
        
end procedure.


/***************************************************************************
*                     PROCEDURE VALIDATE-BY-FIELDS
* 
* Alert if any entered field is not a valid field of the query buffer
*  thereby reducing the potential number of sort matches.
* Ignore BY keyword.
***************************************************************************/

procedure validate-by-fields:
  def input param p-fields as char.

  def var i as int.
  def var by-field as char.
  def var clean-fields as char.
  def var num as int.
  def var ok as log.
  
  if by-fields = p-fields then
    return.
    
  p-fields = clean-whitespace( p-fields ).

  /* CHECK EACH TOKEN OF P-FIELDS */
  do i = 1 to num-entries( p-fields, " " ):
    by-field = entry( i, p-fields, " " ).
    if valid-field( by-field ) then
      assign
        clean-fields = 
          trim( clean-fields + " " +
                entry( num-entries( by-field, "." ), by-field, "." ) )
        num = num + 1.
    else
    if by-field <> "BY" then
    do:
      message 
        by-field "is not a field of" table-name "." skip
        "Only" num "fields are valid for sort matching." skip
        "Are you sure?"
        view-as alert-box question buttons yes-no update ok.
      if not ok then
        return ?.
      leave.
    end. 
  end.
  
  assign
    by-fields = clean-fields
    by-fields:screen-value in frame f = by-fields.
    
  return.
  
end procedure.


/***************************************************************************
*                      PROCEDURE VALIDATE-V6-QUERY
* 
* Make sure v6-query doesn't contradict index-use.
***************************************************************************/

procedure validate-v6-query:
  def input param p-log as char.
  
  if p-log = string( v6-query ) then
    return.
    
  if p-log = "no" and index-use <> "" then
    assign    
      index-use = ""
      index-use:screen-value in frame f = "".
      
  v6-query = (p-log = "yes").
  return.
  
end procedure.


/***************************************************************************
*                     PROCEDURE VALIDATE-INDEX-USE
* 
* Check that entered index-use is a valid index of table-name or blank (no
*  USE-INDEX option).
* Also make sure that this is consistent with v6-query (USE-INDEX implies
*  v6-query = yes).
***************************************************************************/

procedure validate-index-use:
  def input param p-name as char.
  
  def var ok as log.
  
  if index-use = p-name then
    return.
    
  if not (p-name = "" or
          can-find( t-index where 
            t-index.table-name = table-name and
            t-index.index-name = p-name ))
  then
    return p-name + " is not an index of " + table-name.
  
  if p-name <> "" and v6-query = no then
  do:
    message "Specifying an index requires V6 query. Are you sure?"
            view-as alert-box buttons yes-no update ok.
    if not ok then 
      return ?.
    assign
      v6-query = yes
      v6-query:screen-value in frame f = "yes".
  end.
  
  index-use = p-name.
  return.
  
end procedure.


/***************************************************************************
*                        PROCEDURE PARSE-TABLE-DEF
* 
* Attempts to parse the table definition as best it can. No errors are
*  given. Anything not understood is omitted. User is warned about more
*  serious omissions.
* Schema data is created in accordance with interpretation.
* More information about what is expected can be read in ON HELP trigger.
***************************************************************************/

procedure parse-table-def:
  def input param p-text as char.
  
  def var pos as int init 1.
  def var token as char.
  def var mode as char.
  def var index-order as int.
  def var field-order as int.
  def var is-ascending as log.
  def var primary-list as char.
  def var warning as char.
  def var save-dictdb as char.
  
  def buffer t-table for t-table.
  def buffer t-field for t-field.
  def buffer t-index for t-index.
  def buffer t-ixfld for t-ixfld.
  
  /* EXPECT T-TABLE RECORD ALREADY EXISTS */
  find t-table where t-table.table-name = table-name.
  
  /* DELETE ANY EXISTING SCHEMA DATA RELATED TO TABLE */
  for each t-field where t-field.table-name = table-name:
    delete t-field.
  end.
  for each t-index where t-index.table-name = table-name:
    delete t-index.
  end.
  for each t-ixfld where t-ixfld.table-name = table-name:
    delete t-ixfld.
  end.

  /* ADD EXTRA SPACES FOR LOOKAHEAD TOKENS IN CASE INCOMPLETE */
  p-text = p-text + "  ".
  
  /* PROCESS OPTIONAL "LIKE" OPTION */
  if p-text begins "like " then
  do:
    pos = 3.
    run parse-table-like(
      input entry( 2, p-text, " " ),
      output primary-list,
      output index-order
      ).
  end.
  
  /* EXAMINE EACH TOKEN UNTIL NONE LEFT */
  do while pos <= num-entries( p-text, " " ):
    assign
      token = entry( pos, p-text, " " )
      pos = pos + 1.
    
    if token = "" then 
      next.  
    
    /* IF START OF "FIELD" PHRASE */  
    if token = "field" then
    do:
      assign
        token = entry( pos, p-text, " " )
        pos = pos + 1.
      if can-find( t-field where 
          t-field.table-name = table-name and
          t-field.field-name = token )
      then 
        warning = warning + "Field " + token + " already defined.~n".
      else do:
        create t-field.
        assign
          t-field.table-name = table-name
          t-field.field-name = token
          t-field.data-type = "integer"
          mode = "field".
      end.
    end.
    
    /* IF START OF "INDEX PHRASE */
    else
    if token = "index" then
    do:
      if not can-find( first t-field where t-field.table-name = table-name )
      then do:
        warning = warning + "Index specified without fields.~n".
        leave.
      end.
      assign         
        token = entry( pos, p-text, " " )
        pos = pos + 1.
      if can-find( t-index where 
           t-index.table-name = table-name and
           t-index.index-name = token )
      then
        warning = warning + "Index " + token + " already defined.~n".
      else do:
        create t-index.
        assign
          t-index.table-name = table-name
          t-index.index-name = token
          t-index.index-type = "N"
          t-index.index-order = ?
          is-ascending = yes
          mode = "index"
          field-order = 0.
        if not index-by-alpha then
          assign 
            index-order = index-order + 1
            t-index.index-order = index-order.
      end.
    end.
    
    /* ACCEPT "AS" AND "EXTENT" OPTIONS IN FIELD PHRASE */
    else 
    if mode = "field" and available t-field then
    do:
      if token = "as" then
        assign
          t-field.data-type = entry( pos, p-text, " " )
          pos = pos + 1.
      else
      if token = "extent" then
        assign
          t-field.extent = integer( entry( pos, p-text, " " ) )
          pos = pos + 1
          no-error.
    end.

    /* CHECK IF INDEX PHRASE HAS "IS" SUB-PHRASE */
    else
    if mode = "index" and available t-index then
    do:
      if token = "is" or token = "as" then
        mode = "is".
      else
        assign
          mode = "ixfld"
          pos = pos - 1.
    end.
    
    /* "IS" CAN BE "PRIMARY", "UNIQUE", OR "WORD-INDEX" */
    else
    if mode = "is" and available t-index then
    do:
      case token:
        when "primary" then
          primary-list = 
            trim( primary-list + "," + t-index.index-name, "," ).
        when "unique" or when "word-index" then
          if t-index.index-type = "N" then
            t-index.index-type = substring( token, 1, 1 ).
        when "is" or when "as" then /* ignore */.
        otherwise
          assign
            mode = "ixfld"
            pos = pos - 1.
      end case.
    end.
    
    /* IF INDEX-FIELD PHRASE */
    else
    if mode = "ixfld" then
    do:
      if lookup( keyword( token ), "ascending,descending" ) > 0 then
      do:
        if available t-ixfld then
          assign
            is-ascending = token begins "asc"
            t-ixfld.is-ascending = is-ascending.
      end.
      else do:
        find 
          t-field where  
          t-field.table-name = table-name and
          t-field.field-name = token
          no-error.
        if available t-field and t-field.extent = 0 and 
           (t-index.index-type <> "W" or t-index.num-fields = 0)
        then do:
          create t-ixfld.
          assign
            t-ixfld.table-name = table-name
            t-ixfld.index-name = t-index.index-name
            t-ixfld.field-name = token
            field-order = field-order + 1
            t-ixfld.field-order = field-order
            t-ixfld.is-ascending = is-ascending
            t-index.num-field = t-index.num-fields + 1
            t-ixfld.is-word = t-index.index-type = "W".
        end.
        else
          warning = 
            warning +
            (if not available t-field then
               "Index field " + token + " is not a defined field.~n"
             else
             if t-field.extent > 0 then
               "Index field " + token + " is an array.~n"
             else "").
      end.
    end.   
          
  end. /* do while tokens */

  /* TIDY DEFINITION DATA AFTER TEXT HAS BEEN PARSED */
  run parse-table-tidy (
    buffer t-table,
    input primary-list,
    input warning
    ).
           
end procedure.


/***************************************************************************
*                     PROCEDURE PARSE-TABLE-LIKE
* 
* Copy schema from p-like-table to table-name.
* If p-like-table is unqualified database table then look for unique table
*  with this name across all connected databases.
* Assign statements are used rather than buffer-compare for earlier
*  versions.
***************************************************************************/

procedure parse-table-like:
  def input param p-like-table as char.
  def output param p-primary-list as char.
  def output param p-index-order as int.

  def buffer t-table for t-table.
  def buffer t-field for t-field.
  def buffer t-index for t-index.
  def buffer t-ixfld for t-ixfld.
  def buffer like-table for t-table.
  def buffer like-field for t-field.
  def buffer like-index for t-index.
  def buffer like-ixfld for t-ixfld.

  /* IGNORE "LIKE" OPTION IF CAN'T FIND UNIQUE TABLE FOR P-LIKE-TABLE */ 
  find like-table where like-table.table-name = p-like-table no-error.
  if not available like-table and index( p-like-table, "." ) = 0 then
    find like-table where like-table.file-name = p-like-table no-error.
  if not available like-table then
    return.

  assign
    p-like-table = like-table.table-name    
    p-primary-list = like-table.primary-index.
    
  /* COPY FIELDS */  
  for each like-field where like-field.table-name = p-like-table:
    create t-field.
    assign
      t-field.table-name = table-name
      t-field.field-name = like-field.field-name
      t-field.data-type = like-field.data-type
      t-field.extent = like-field.extent.
  end.
  
  /* COPY INDEXES */
  for each like-index where like-index.table-name = p-like-table:
    create t-index.
    assign
      t-index.table-name = table-name
      t-index.index-name = like-index.index-name
      t-index.index-order = ?
      t-index.index-type = like-index.index-type
      t-index.num-fields = like-index.num-fields.
    if db-name = "" then 
      assign
        p-index-order = p-index-order + 1
        t-index.index-order = p-index-order.
  end.
  
  /* COPY INDEX FIELDS */
  for each like-ixfld where like-ixfld.table-name = p-like-table:
    create t-ixfld.
    assign
      t-ixfld.table-name = table-name
      t-ixfld.index-name = like-ixfld.index-name
      t-ixfld.field-name = like-ixfld.field-name
      t-ixfld.field-order = like-ixfld.field-order
      t-ixfld.is-ascending = like-ixfld.is-ascending.
  end.

end procedure.


/***************************************************************************
*                     PROCEDURE PARSE-TABLE-TIDY
* 
* An extension of "parse-table-def".
* After the definition text has been parsed, perform the following tasks:
* - If no fields have been defined then return the defintion to its original
*  state (if existing, otherwise delete it).
* - Delete any indexes that have no fields.
* - If no non-word indexes are defined, create a default index.
* - Make sure one and only one non-word index is primary (the first
*  index defined as primary or the first index defined).
* - If order-by-alpha then reorder indexes alphabetically by name.
* - Message alert any warnings that have accrued (there will be no errors).
***************************************************************************/

procedure parse-table-tidy:
  def param buffer t-table for t-table.
  def input param p-primary-list as char.
  def input param p-warning as char.
  
  def var pos as int.
  def var save-dictdb as char.
  def var index-order as int.
  
  /* IF NO FIELDS THEN RETURN DEFINITION TO ORIGINAL STATE */
  if not can-find( first t-field where t-field.table-name = table-name )
  then do:
    p-warning = 
      p-warning + "Table has no defined fields.~n" +
      (if t-table.from-db = no then "Deleting table.~n"
       else "Returning to original DB table definition.~n").
    if t-table.from-db then
    do:
      delete t-table.
      save-dictdb = ldbname( "dictdb" ).
      create alias dictdb for database value( db-name ).
      run qidx-db.p ( db-name, file-name ).
      create alias dictdb for database value( db-name ).
    end.
    else
      delete t-table.
  end.

  else do:
   
    /* DELETE ANY INDEXES THAT HAVE NO FIELDS */
    for each t-index where 
             t-index.table-name = table-name and
             t-index.num-fields = 0:
      assign
        p-warning = 
          p-warning + "Index " + t-index.index-name + " has no fields.~n"
        pos = lookup( t-index.index-name, p-primary-list ).
      if pos > 0 then
        entry( pos, p-primary-list ) = "".
      delete t-index.
    end.
  
    /* IF NO NON-WORD INDEXES THEN CREATE DEFAULT INDEX */
    if not can-find( first t-index where 
             t-index.table-name = table-name and
             t-index.index-type <> "W" )
    then do:
      create t-index.
      create t-ixfld.
      assign
        t-index.table-name = table-name
        t-index.index-name = "default"
        t-index.index-order = 1
        t-index.index-type = "D"
        t-index.num-fields = 1
        t-ixfld.table-name = table-name
        t-ixfld.index-name = t-index.index-name
        t-ixfld.field-name = ""
        t-ixfld.field-order = 1
        t-ixfld.is-ascending = yes
        t-table.primary-index = t-index.index-name.
    end.
    
    else do:
    
      /* MAKE SURE ONE AND ONLY ONE PRIMARY INDEX */
      p-primary-list = trim( p-primary-list, "," ).
      for first t-index where
                t-index.table-name = table-name and
                t-index.index-type <> "W" and
                (lookup( t-index.index-name, p-primary-list ) > 0 or
                 p-primary-list = ""):
        t-table.primary-index = t-index.index-name.
      end.
      
      /* IF INDEX-BY-ALPHA, ORDER INDEXES ALPHABETICALLY */
      if index-by-alpha then
        for each t-index where
                 t-index.table-name = table-name
            by t-index.index-name:
          assign
            index-order = index-order + 1
            t-index.index-order = index-order.
        end.
    end.
   
  end.

  if p-warning <> "" then
    message p-warning view-as alert-box warning.
    
end procedure.


/***************************************************************************
*                     PROCEDURE PARSE-WHERE-EXPR
* 
* Parse where-expr and convert text to where-tree data.
* Any text within brackets is parsed first, starting with innermost bracket.
*  The bracketed text is replaced with an escaped wid representing the
*  where-tree data this text was converted to (this will be processed later
*  in "analyse-where-escaped").
* The output of the outermost (implicit) bracket parse will be the wid of 
*  the root node of the where-tree.
* A where-tree is created even if where-expr is blank (it will be
*  interpreted like any other non-indexable expression and will lead to a
*  whole-index being selected).
* This parsing procedure is designed to be relatively simple but functional
*  enough to interpret most valid where expressions as expected. There are
*  a number of limitations, some of which are outlined in "ON HELP" trigger.
***************************************************************************/

procedure parse-where-expr:
  def input param p-text as char.
  
  def var r-pos as int.
  def var l-pos as int.
  def var new-wid as int.

  /* CLEAN WHITESPACE AND IGNORE "WHERE" KEYWORD IF INCLUDED */
  p-text = clean-whitespace( p-text ).
  if p-text begins "WHERE " then
    p-text = substring( p-text, 7 ).
  
  /* PROCESS ALL BRACKETED PORTIONS OF TEXT FIRST */
  do while index( p-text, ")" ) > 0:
  
    /* FIND POSITIONS OF LEFTMOST INNER BRACKET */
    assign
      r-pos = index( p-text, ")" )
      l-pos = r-index( substring( p-text, 1, r-pos ), "(" ).

    /* PARSE TEXT INSIDE BRACKETS AND REPLACE WITH ESCAPED WID */
    run parse-where-bracket( 
      input trim( substring( p-text, l-pos + 1, r-pos - l-pos - 1 ) ),
      input ?,
      output new-wid
      ).
    assign
      substring( p-text, l-pos, r-pos - l-pos + 1 ) = 
        " ~E" + string( new-wid ) + " "
      p-text = clean-whitespace( p-text ).

  end.
  
  /* PARSE WHOLE TEXT */
  run parse-where-bracket( 
    input p-text,
    input 0,
    output root-wid
    ).
          
end procedure.


/***************************************************************************
*                    PROCEDURE PARSE-WHERE-BRACKET
* 
* Parse bracketed text by dividing text into OR components (separated by
*  " OR ") and parse each component.
* If there is more than one component then create an OR where-tree node to
*  attach children to.
***************************************************************************/

procedure parse-where-bracket:
  def input param p-text as char.
  def input param p-parent-wid as int.
  def output param p-new-wid as int.

  def var i as int.
  def var parent-wid as int.
  def var new-wid as int.
  def var word-index as char.
  def var child-num as int.
  def var pos as int.
  
  def buffer t-wtree for t-where-tree.
  
  /* CREATE "OR" NODE IF MORE THAN ONE COMPONENT */
  pos = index( p-text, " OR " ).
  if pos > 0 then
  do:
    create t-wtree.
    assign
      wid = wid + 1
      t-wtree.child-num = next-wtree-child( p-parent-wid )
      t-wtree.wid = wid
      t-wtree.inode = wid
      t-wtree.parent-wid = p-parent-wid
      t-wtree.type = "OR"
      t-wtree.i-state = "I"
      p-parent-wid = wid
      p-new-wid = wid.
  end.
      
  /* PROCESS EACH COMPONENT */      
  repeat:
    run parse-where-or (
      input (if pos = 0 then p-text else substring( p-text, 1, pos - 1 )),
      input p-parent-wid,
      output new-wid,
      output word-index
      ).
    if not available t-wtree then
      p-new-wid = new-wid.
    else
    if word-index = "*" then
      t-wtree.word-index = "*".
    if pos = 0 then
      leave.
    assign
      p-text = substring( p-text, pos + 4 )
      pos = index( p-text, " OR " ).
  end.

end procedure.


/***************************************************************************
*                      PROCEDURE PARSE-WHERE-OR
* 
* Parse OR component by dividing text into AND components (separated by
*  " AND "), and parse each AND component.
* If there is more than one AND component then create an AND where-tree node
*  to attach children to.
***************************************************************************/

procedure parse-where-or:
  def input param p-text as char.
  def input param p-parent-wid as int.
  def output param p-new-wid as int.
  def output param p-word-index as char.

  def var i as int.
  def var new-wid as int.
  def var word-index as char.
  def var child-num as int.
  def var pos as int.
    
  def buffer t-wtree for t-where-tree.
    
  /* CREATE "AND" NODE IF MORE THAN ONE COMPONENT */
  pos = index( p-text, " AND " ).
  if pos > 0 then
  do:
    create t-wtree.
    assign
      wid = wid + 1
      t-wtree.child-num = next-wtree-child( p-parent-wid )
      t-wtree.wid = wid
      t-wtree.parent-wid = p-parent-wid
      t-wtree.inode-wid = wid
      t-wtree.type = "AND"
      t-wtree.i-state = "I"
      p-parent-wid = wid.
  end.
      
  /* PROCESS EACH COMPONENT */      
  repeat:
    run parse-where-single (
      input (if pos = 0 then p-text else substring( p-text, 1, pos - 1 )),
      input p-parent-wid,
      input available t-wtree,
      output new-wid,
      output word-index
      ).
    if word-index = "*" then
      p-word-index = "*".      
    if pos = 0 then
      leave.
    assign
      p-text = substring( p-text, pos + 5 )
      pos = index( p-text, " AND " ).
  end.
  
  if available t-wtree then
    assign
      t-wtree.word-index = p-word-index
      p-new-wid = p-parent-wid.
  else
    p-new-wid = new-wid.

end procedure.


/***************************************************************************
*                     PROCEDURE PARSE-WHERE-SINGLE
* 
* Parses a single phrase of the where expression that doesn't include any
*  AND or OR operators.
* Any escaped wids in the text are processed first. If "parse-where-escaped"
*  outputs empty p-text then the where-tree data that was already created
*  for the bracketed text represented by the escaped wid is attached to the
*  parent and no further processing is needed here.
* Otherwise a where-tree leaf is created and p-text is analysed to see if it
*  may be useful for indexing. It may be if it is of the form: 
*  "<field> <operator> <expression>" - where <field> is an indexed field of
*  the query buffer, <operator> is a valid operator (listed in operators
*  variable); or "<function> ( <buffer> ) <operator> <expression>" - where
*  <function> is "RECID" or "ROWID", <buffer> is the query buffer,
*  <operator> is "=" or "EQ". In both cases, <expression> can be almost
*  anything - this isn't parsed except to look for a field reference that
*  would exclude it from indexing.
* Note that if a RHS expression includes a field reference within a UDF
*  then it will be incorrectly excluded from indexing. This parser isn't
*  able to recognise a UDF, so any field reference in a UDF needs be
*  modified to avoid an erroneous result.
* If a phrase is not useful for indexing then an explanatory comment and
*  e-state of "X" will be assigned.
* Otherwise, "parse-where-e-state" will be run to select the right e-state. 
*  And if word-index then find the first word-index for field.
***************************************************************************/

procedure parse-where-single:
  def input param p-text as char.
  def input param p-parent-wid as int.
  def input param p-and as log.
  def output param p-new-wid as int.
  def output param p-word-index as char.
  
  def var field-name as char.
  def var operator as char.
  def var op-pos as int.
  def var state as char.
  def var comment as char.
  def var i as int.
  def var operators as char init  
    "=,EQ,BEGINS,>=,GE,>,GT,<=,LE,<,LT,CONTAINS,RECID".
  def var state-nums as int extent 13 init [1,1,2,3,3,3,3,4,4,4,4,5,6].

  def buffer t-wtree for t-where-tree.
   
  /* PROCESS ANY ESCAPED WIDs THAT ARE STILL IN THE TEXT */
  if index( p-text, "~E" ) > 0 then
  do:
    run parse-where-escaped( 
      input-output p-text,
      input p-parent-wid,
      output p-new-wid,
      output p-word-index
      ).
    if p-text = "" then
      return.
  end.
  
  /* CREATE WTREE RECORD FOR SINGLE PHRASE */
  create t-wtree.
  assign
    t-wtree.child-num = next-wtree-child( p-parent-wid )
    wid = wid + 1
    t-wtree.wid = wid
    t-wtree.parent-wid = p-parent-wid
    t-wtree.inode-wid = (if p-and then p-parent-wid else wid)
    t-wtree.type = "LEAF"
    t-wtree.i-state = (if p-and then "<" else "I")
    t-wtree.expr = p-text
    p-new-wid = wid.

  /* INTERPRET SINGLE PHRASE */
  if num-entries( p-text, " " ) >= 3 then
  do:
    field-name = entry( 1, p-text, " " ).
    
    /* FIRST CHECK IF VALID RECID/ROWID PHRASE */
    if lookup( field-name, "recid,rowid" ) > 0 and
       num-entries( p-text, " " ) > 5 and
       entry( 2, p-text, " " ) = "(" and
       valid-field( entry( 3, p-text, " " ) + ".*" ) and
       entry( 4, p-text, " " ) = ")" and
       lookup( entry( 5, p-text, " " ), "=,EQ" ) > 0
    then
      assign
        field-name = "RECID"
        op-pos = lookup( field-name, operators ).
    
    /* NEXT CHECK IF VALID FIELD-NAME AND OPERATOR */
    else do:
      assign
        operator = entry( 2, p-text, " " )
        op-pos = lookup( operator, operators ).
    
      /* CHECK IF NON-INDEXABLE OPERATOR */ 
      if op-pos = 0 or operator = "recid" then
        comment = "non-indexable Operator".
      
      /* CHECK IF INVALID FIELD OF BUFFER */
      else
      if not valid-field( field-name ) then
        comment = "non-buffer-field".
      
      /* CHECK IF FIELD-NAME IS PROPERLY INDEXED */
      else do:
      
        /* STRIP FIELD-NAME OF ANY BUFFER QUALIFIER */
        field-name = 
          entry( num-entries( field-name, "." ), field-name, "." ).
          
        /* CHECK IF FIELD IS NOT PART OF ANY INDEX */
        if not can-find( first t-ixfld where
                 t-ixfld.table-name = table-name and
                 t-ixfld.field-name = field-name and
                 t-ixfld.is-word = (operator = "CONTAINS") )
        then
          comment = 
            "field not " + (if operator = "CONTAINS" then "word-" else "") +
            "indexed".
          
        /* DON'T RECOGNISE CONTAINS IF V6-QUERY */
        else
        if operator = "CONTAINS" and v6-query then
          comment = "V6 excludes CONTAINS".
                 
      end. /* else valid field */

    end. /* else not recid */
  
    /* IF EXPR INCLUDES REFERENCE TO VALID FIELD THEN NOT INDEXABLE */
    if comment = "" then
      do i = (if field-name = "RECID" then 6 else 3) to 
             num-entries( p-text, " " ):
        if valid-field( entry( i, p-text, " " ) ) then
        do:
          comment = "RHS field reference".
          leave.
        end.
      end.
      
  end.

  else 
    comment = 
      (if p-text = "" then "no expression"
       else "non-operator phrase").
    
  /* IF NO COMMENT YET THEN POTENTIALLY INDEXABLE E-STATE */
  if comment = "" then
  do:
    t-wtree.field-name = field-name.

    /* RUN FINAL TESTS TO DETERMINE E-STATE */
    run parse-where-e-state(
      buffer t-wtree,
      input state-nums[ op-pos ],
      input p-and
      ).
    
    /* FIND RIGHT INDEX IF WORD-INDEX */  
    if t-wtree.e-state = "IW" then 
      for each t-index where 
               t-index.table-name = table-name and
               t-index.index-type = "W",
          each t-ixfld where
               t-ixfld.table-name = table-name and
               t-ixfld.index-name = t-index.index-name and
               t-ixfld.field-name = field-name
          by t-index.index-order:
        assign
          t-wtree.word-index = t-index.index-name
          t-wtree.comment = t-wtree.comment + " " + t-wtree.word-index
          p-word-index = "*".
        leave.
      end.        
          
  end.
  
  /* ELSE EXCLUDED */
  else
    assign
      t-wtree.e-state = "X"
      t-wtree.comment = comment.
  
end procedure.


/***************************************************************************
*                    PROCEDURE PARSE-WHERE-ESCAPED
* 
* If "~E" character exists in p-text it means that internal bracketed text
*  has already been converted into where-tree data and replaced by escaped
*  wid. Now it is time to process this.
* If the entire text consists of an escaped wid then the bracket data
*  will be attached to p-parent-wid. If bracket node has same operator as
*  parent then children of bracket are attached to parent and bracket node
*  is deleted (Progress doesn't seem to give precedence to bracketed
*  phrases). Otherwise bracket node is attached to parent.
* If there is more to p-text than a single escaped wid then any bracket data
*  has to be converted back into text to be considered as part of the single
*  phrase without any special meaning (e.g. "field = (a OR b)"). The
*  converted text may be a little different from the original. Brackets
*  will always be added with spaces around them - parsing of RECID() expects
*  this in "parse-where-single".
***************************************************************************/

procedure parse-where-escaped:
  def input-output param p-text as char.
  def input param p-parent-wid as int.
  def output param p-new-wid as int.
  def output param p-word-index as char.
  
  def var bracket-wid as int.
  def var parent-type as char.
  def var child-num as int.
  def var l-pos as int.
  def var r-pos as int.
  def var bracket-text as char.
   
  def buffer t-wtree for t-where-tree.
  
  /* IF ENTIRE TEXT IS ESCAPED THEN ATTACH BRACKET DATA TO PARENT */
  if p-text begins "~E" and index( p-text, " " ) = 0 then
  do:
    bracket-wid = integer( substring( p-text, 2 ) ) no-error.
    if bracket-wid > 0 then
    do:
      find t-wtree where t-wtree.wid = p-parent-wid no-error.
      parent-type = (if available t-wtree then t-wtree.type else ?).
      find t-wtree where t-wtree.wid = bracket-wid no-error.
      if available t-wtree and t-wtree.parent-wid = ? then
      do:
        child-num = next-wtree-child( p-parent-wid ).
       
        /* IF TYPE SAME AS PARENT THEN CHILDREN SKIP A GENERATION */
        if t-wtree.type = parent-type then
        do:
          delete t-wtree.
          for each t-wtree where
                   t-wtree.parent-wid = bracket-wid:
            assign
              t-wtree.parent-wid = p-parent-wid
              t-wtree.child-num = child-num
              t-wtree.inode-wid = 
                (if parent-type = "AND" then p-parent-wid else t-wtree.wid)
              child-num = child-num + 1.
          end.
        end.
        
        /* ELSE ATTACH BRACKET AS CHILD TO PARENT */
        else
          assign
            t-wtree.parent-wid = p-parent-wid
            t-wtree.child-num = child-num
            t-wtree.inode-wid = 
              (if parent-type = "AND" then p-parent-wid else t-wtree.wid)
            p-new-wid = t-wtree.wid
            p-word-index = t-wtree.word-index.
      
      end.
    end.
    p-text = "".
    return.
  end.
  
  /* OTHERWISE CONVERT ALL ESCAPED DATA BACK INTO TEXT AS BEST WE CAN */ 
  repeat:  
    assign
      bracket-wid = 0
      l-pos = index( p-text, "~E" ).
    if l-pos = 0 then
      return.
    assign
      r-pos = index( p-text + " ", " ", l-pos )
      bracket-wid = 
        integer( substring( p-text, l-pos + 1, r-pos - l-pos - 1 ) )
      no-error.
    run report-where-tree(
      input bracket-wid,
      input no,
      input "convert",
      input 1,
      output bracket-text
      ).
    substring( p-text, l-pos, r-pos - l-pos ) = 
      "( " + bracket-text + " )".
  end.
      
end procedure.
 
 
/***************************************************************************
*                     PROCEDURE PARSE-WHERE-E-STATE
*
* This procedure defines the indexable expression states and comments for a
*  where-tree leaf.
* Procedure "parse-where-single" has already determined the state-num of the
*  indexable e-state it could potentially be based on the field-name and
*  operator, but now older siblings need to be checked.
* Each e-state has its own precede-list and override-list of e-states.
* A Word-Index does not precede or override anything. Each other e-state has
*  its own e-state and higher e-states on its precede-list, and lower
*  e-states on its override-list.
* The lists for Range-Both depend on the existence of both Range-Left and
*  and Range-Right siblings. If both exist then Range-Left and Range-Right
*  are added to the precede-list, otherwise they are added to the
*  override-list.
* An older sibling with the same field-name and its e-state on the
*  precede-list will precede and exclude the current leaf.
* An older sibling with the same field-name and its e-state on the
*  override-list will be overridden and excluded by the current leaf.
***************************************************************************/

procedure parse-where-e-state:
  def param buffer pt-wtree for t-where-tree.
  def input param p-state-num as int.
  def input param p-and as log.
  
  def var wnode as char.
  def var state as char.
  def var precede-list as char.
  def var override-list as char.
  def var states as char init "IEQ,IRB,IRL,IRR,IW,IID".
  def var comments as char extent 6 init[
    "EQUALITY", 
    "RANGE-BOTH",
    "RANGE-LEFT",
    "RANGE-RIGHT",
    "WORD-INDEX",
    "RECID/ROWID"
    ].
  def var overrides as char init "IRB,IRL,IRR;IRL,IRR;;;;".
  def var precedes as char init "IEQ;IEQ,IRB;IRL,IRB,IEQ;IRR,IRB,IEQ;;IID".
   
  def buffer t-wtree for t-where-tree.
  
  /* ASSIGN STATE AND LISTS BASED ON P-STATE-NUM */
  assign
    state = entry( p-state-num, states )
    precede-list = entry( p-state-num, precedes, ";" )
    override-list = entry( p-state-num, overrides, ";" ).

  /* IF AND-PARENT THEN CHECK SIBLINGS FOR PRECEDING AND OVERRIDING */
  if p-and then
  do:
  
    /* IF STATE IS "IRB" THEN "IRL" & "IRR" PAIR WILL AFFECT LISTS */
    if state = "IRB" then
      if can-find( first t-wtree where
           t-wtree.parent-wid = pt-wtree.parent-wid and
           t-wtree.field-name = pt-wtree.field-name and
           t-wtree.e-state = "IRL" ) and
         can-find( first t-wtree where
           t-wtree.parent-wid = pt-wtree.parent-wid and
           t-wtree.field-name = pt-wtree.field-name and
           t-wtree.e-state = "IRR" )
      then
        precede-list = precede-list + ",IRL,IRR".
      else
        override-list = override-list + ",IRL,IRR".
    
    /* CHECK IF PRECEDED BY AN OLDER SIBLING */
    if precede-list <> "" then
      for each t-wtree where 
               t-wtree.parent-wid = pt-wtree.parent-wid and
               t-wtree.field-name = pt-wtree.field-name and
               lookup( t-wtree.e-state, precede-list ) > 0:
        wnode = wnode + "&[" + string( t-wtree.wid, "99" ) + "]".    
        if wnode begins "&" then
          assign
            pt-wtree.field-name = ""
            state = "X"
            pt-wtree.comment = "preceded by " + substring( wnode, 2 ). 
      end.
    
    /* CHECK IF OVERRIDE ANY OLDER SIBLINGS */
    if override-list <> "" and pt-wtree.e-state <> "X" then
      for each t-wtree where
               t-wtree.parent-wid = pt-wtree.parent-wid and
               t-wtree.field-name = pt-wtree.field-name and
               lookup( t-wtree.e-state, override-list ) > 0:
        assign
          t-wtree.field-name = ""
          t-wtree.e-state = "X"
          t-wtree.comment = 
            "overridden by [" + string( pt-wtree.wid, "99" ) + "]".
      end.

  end. /* if p-and */
  
  /* FINALISE E-STATE AND COMMENT */
  pt-wtree.e-state = state.
  if state <> "X" then
    pt-wtree.comment = comments[ p-state-num ].
        
end procedure.


/***************************************************************************
*                     PROCEDURE ANALYSE-WHERE-TREE
* 
* The where-tree is analysed recursively to determine which nodes of its
*  structure are indexable, which indexes the compiler would select for
*  these nodes, and the index-tree structure the compiler would build.
* Any where-tree node that is passed to this procedure for analysis is
*  considered to be an indexable node (inode) if it is not an OR-node or
*  V6-query = yes (in which case the root node is always indexable). An
*  inode will have one or more indexes selected for it unless it ends up
*  being excluded for some reason.
* A tricky aspect of analysing an AND-node is that non-FEM indexes will be
*  excluded from selection if OR components are included but it is not
*  always known if an OR component will be included until it is known if the
*  inode would be whole-index without them. Therefore, indexes are analysed
*  for single phrases first, then it is determined if the inode would be
*  whole-index, then children (just OR-children for an AND-node) are
*  processed, and then index selection is completed.
* A distinction is made here between true-whole and prog-whole. An inode is
*  truly whole-index if it has no word-index components, no OR components
*  (initially unknown), all test indexes have no equality or range matches,
*  and there are no RECID/ROWID searches. But if it is an AND-inode and any
*  index has a sort match and there is an equality or range search on any
*  indexable field then Progress for some unknown reason deems it to be not
*  "whole-index", even though it may well be whole-index if the indexable
*  field is not the first component of any index. It is this Progress
*  determination of whole-index that has significant influence over index
*  selection so prog-whole is the version of whole-index that we must deal
*  with here.
* If V6-query then the index-tree will just be a single leaf linked to the
*  root node of the where-tree, no matter what the where expression, so
*  the where-tree is not analysed beyond the root node.
* Leaf children of an AND-node are not included in this analysis because
*  any indexes will be attached to the AND-node, not the leaf. But
*  leaf-children of an OR-node are individual inodes so these are included.
* An OR-node that is the child of an AND-node will be excluded from analysis
*  unless it has a word-index component (in which case it will always be
*  included) or the AND-node was deemed to be whole-index without it (in
*  which case only the first OR-node child that doesn't have a whole-index
*  child of its own will be included).
* Any child of an OR-node that is known to have a whole-index sibling will
*  be excluded from this analysis unless it has a word-indexed component.
* Indexes are selected for an inode after OR-children have been processed.
* Finally "analyse-where-index" is run to shape the index-tree based on the
*  analysis of this node of the where-tree.
***************************************************************************/

procedure analyse-where-tree:
  def input param p-wid as int.
  def input param p-parent-type as char.
  def input param p-parent-iid as int.
  
  def var node-iid as int.
  def var is-inode as log.
  def var is-and-node as log.
  def var true-whole as log.
  def var prog-whole as log.
  def var accept-or-wid as int.
  
  def buffer t-wtree for t-where-tree.
  def buffer child-wtree for t-where-tree.
  def buffer gchild-wtree for t-where-tree.
  def buffer t-tindex for t-test-index.
  
  find t-wtree where t-wtree.wid = p-wid no-error.
  if not available t-wtree then
    return.

  /* RESERVE NEW IID TO ATTACH ANY NEW INDEX-TREE CHILDREN TO */
  assign
    iid = iid + 1
    node-iid = iid
    is-and-node = t-wtree.type = "AND"
    is-inode = v6-query or t-wtree.type <> "OR".
  
  /* IF INODE THEN ANALYSE INDEXES AND DETERMINE IF "WHOLE-INDEX" */
  if is-inode then
  do:
    run analyse-indexes( 
      input t-wtree.wid,
      input is-and-node
      ).
   
    /* REAL WHOLE-INDEX TEST (DOESN'T AFFECT INDEX SELECTION) */
    true-whole =
      t-wtree.word-index = "" and 
      not can-find( first t-tindex where
            t-tindex.inode-wid = t-wtree.wid and
            (t-tindex.fem <> "" or
             t-tindex.num-equality > 0 or
             t-tindex.num-range > 0 ) ).
    
    /* WHOLE-INDEX TEST THAT ACTUALLY AFFECTS INDEX SELECTION */
    prog-whole =
      true-whole and
      (if t-wtree.type = "AND" then
         not (can-find( first t-tindex where
                t-tindex.inode-wid = t-wtree.wid and
                t-tindex.num-sort > 0 ) and
              can-find( first child-wtree where
                child-wtree.parent-wid = t-wtree.wid and
                child-wtree.e-state begins "I" ))
       else yes).

  end.
  
  /* PROCESS ALL CHILDREN (EXCEPT LEAVES OF AND-NODE) IF NOT V6-QUERY */ 
  if not v6-query then
    for each child-wtree where 
             child-wtree.parent-wid = t-wtree.wid and
             not (child-wtree.type = "LEAF" and is-and-node):
      
      /* CHECK IF CHILD IS EXCLUDED BY WHOLE-INDEX SIBLING */
      if t-wtree.whole-wid > 0 and child-wtree.word-index = "" then
        assign
          child-wtree.i-state = "X"
          child-wtree.comment = 
            "whole-index [" + string( t-wtree.whole-wid, "99" ) + "]".
            
      /* CHECK IF OR-CHILD EXCLUDED */
      else 
      if child-wtree.type = "OR" and child-wtree.word-index = "" and
         not prog-whole
      then
        assign
          child-wtree.i-state = "X"
          child-wtree.comment = 
            (if accept-or-wid > 0 
             then "preceded by OR [" + string( accept-or-wid, "99" ) + "]"
             else "excluded OR child"). 
      
      /* RECURSIVELY ANALYSE NON-EXCLUDED CHILD */
      else do:
        run analyse-where-tree( 
          input child-wtree.wid,
          input t-wtree.type,
          input node-iid
          ).
  
        /* UNDO OR-CHILD IF WHOLE-INDEX (AND NOT WORD) */
        if prog-whole and child-wtree.whole-wid > 0 then
        do:
          assign
            child-wtree.i-state = "X"
            child-wtree.comment = 
              "whole-index child [" + 
              string( child-wtree.whole-wid, "99" ) + "]".
          for each gchild-wtree where 
                   gchild-wtree.parent-wid = child-wtree.wid:
            run analyse-where-undo(
              input gchild-wtree.wid,
              input 0
              ).
          end.
        end.
          
        /* ELSE ACCEPT OR-CHILD */
        else
          assign
            accept-or-wid = child-wtree.wid
            prog-whole = no.
      
      end. 
                  
    end. /* for each child-wtree */
  
  /* SELECT INDEXES IF INODE */    
  if is-inode then
    run select-indexes(
      input t-wtree.wid,
      input node-iid,
      input is-and-node,
      input prog-whole
      ).
  
  run analyse-where-index(
    buffer t-wtree,
    input p-parent-type,
    input p-parent-iid,
    input node-iid
    ).
    
end procedure.


/***************************************************************************
*                     PROCEDURE ANALYSE-WHERE-INDEX
*
* This is an extension of "analyse-where-tree" to construct and possibly
*  rearrange the index-tree around this node of the where-tree.
* Initially, no index-tree record has been created for p-node-iid - this iid
*  has been reserved for a potential OR-node or AND-node, and any index-tree
*  leaves created in "select-indexes" will be attached as children to this.
* Now it is time to determine if p-node-iid has just one or multiple
*  index-tree children. (A leaf inode will always have a single index-tree
*  child. An AND-inode may have a single child or multiple children,
*  depending on index selection. An OR-node will typically have an
*  index-tree child for each of its where-tree children, but if one child is
*  a whole-index it could easily become an only-child.)
* If the unique FIND reveals an only-child, then some index-tree rearranging
*  will be needed. 
* If the only-child is a whole-index and its parent is an OR-node then all
*  its older siblings without a word-index component need to be undone.
*  Those with a word-index component are renumbered. 
* The only-child is promoted to become the next child of p-parent-iid.
* It is possible that the current where-tree node is an AND-node with an
*  OR-parent and an OR-child that is also an only-child (this is only
*  possible if the OR-child has a word-index descendant and other children
*  of the AND-node don't allow any non-word index to be FEM). In this case,
*  the AND-node has just been replaced by its OR-child giving the OR-child
*  an OR-parent. The next step is to merge the OR-child with the OR-parent.
*  The children of the OR-child are promoted to become children of the
*  OR-parent. However, if the OR-parent already has a whole-index child then
*  any children of the OR-child that don't have a word-index component need
*  to be undone and excluded.
* If the unique FIND reveals more than one child then a new index-tree node
*  is created to be the parent of these children.
* The index-tree rearranging for an only-child may be the most complicated
*  and difficult part of this entire program. The basic rule is that if an
*  index-tree OR-node ends up with a whole-index child then it is not
*  allowed to have any other children except those with a word-index
*  component. It is possible for many generations of OR-nodes to be merged,
*  and the tricky part is determining which where-tree components from
*  different generations will end up being siblings in the index-tree while
*  still being able to comment which node caused another node to be merged,
*  replaced, excluded, etc. That's what this procedure does.
***************************************************************************/

procedure analyse-where-index:
  def param buffer pt-wtree for t-where-tree.
  def input param p-parent-type as char.
  def input param p-parent-iid as int.
  def input param p-node-iid as int.
    
  def var or-iid as int.
  def var child-num as int.
  def var whole-wid as int.
  
  def buffer t-wtree for t-where-tree.
  def buffer parent-wtree for t-where-tree.
  def buffer child-wtree for t-where-tree.
  def buffer t-itree for t-index-tree.
  def buffer child-itree for t-index-tree.
  def buffer t-tindex for t-test-index.
  
  child-num = next-itree-child( p-parent-iid ).
      
  /* CHECK IF NODE-IID ONLY HAS ONE CHILD */
  find t-itree where t-itree.parent-iid = p-node-iid no-error.
  if available t-itree and not ambiguous t-itree then
  do:
   
    /* IF OR-PARENT THEN MAY NEED UNDO OR REARRANGE FAMILY */
    if p-parent-type = "OR" and t-itree.whole-wid > 0 then
    do:
      find parent-wtree where parent-wtree.wid = pt-wtree.parent-wid.
      whole-wid = parent-wtree.whole-wid.
      if whole-wid = 0 then 
      do:
        assign
          parent-wtree.whole-wid = t-itree.whole-wid
          child-num = 1.
        for each child-itree where child-itree.parent-iid = p-parent-iid:
          if child-itree.word-wid = 0 then
            run analyse-where-undo(
              input child-itree.inode-wid,
              input t-itree.whole-wid
              ).
          else
            assign
              child-itree.child-num = child-num
              child-num = child-num + 1.
        end.
      end.
    end.
     
    /* PROMOTE ONLY-CHILD AS A CHILD OF PARENT-IID */
    assign
      p-node-iid = t-itree.iid
      t-itree.parent-iid = p-parent-iid
      t-itree.child-num = child-num.
    if pt-wtree.type = "AND" and t-itree.type = "LEAF" or v6-query then
      pt-wtree.comment = "single index".
    else
    if pt-wtree.type <> "LEAF" then
      assign
        pt-wtree.i-state = "-"
        pt-wtree.comment = 
          "replaced by [" + string( t-itree.orig-wid, "99" ) + "]".
    if pt-wtree.wid = root-wid then
      root-iid = p-node-iid.
    
    /* IF NOW OR-CHILD OF OR-PARENT THEN MERGE */
    if p-parent-type = "OR" and t-itree.type = "OR" then
    do:
    
      /* FIND OR-CHILD WHERE-TREE TO MARK AS MERGED */
      find child-wtree where child-wtree.wid = t-itree.orig-wid.
      assign
        child-wtree.i-state = "-"
        child-wtree.comment = 
          "merged into [" + 
          string( pt-wtree.parent-wid, "99" ) + "]"
        or-iid = t-itree.iid.
        
      /* UNDO INDEX ANALYSIS FOR AND-NODE (ALREADY REPLACED BY OR-CHILD) */  
      run analyse-where-undo(
        input pt-wtree.wid,
        input 0
        ).
      delete t-itree.

      /* PROMOTE OR UNDO EACH CHILD OF OR-CHILD */
      for each child-itree where child-itree.parent-iid = or-iid:
     
        /* IF PREVIOUS WHOLE-INDEX AND NON-WORD THEN UNDO */
        if whole-wid > 0 and child-itree.word-wid = 0 then
          run analyse-where-undo(
            input child-itree.inode-wid,
            input whole-wid
            ).
            
        /* ELSE PROMOTE GRANDCHILD TO HIGHER GENERATION */    
        else
          assign
            child-itree.parent-iid = p-parent-iid
            child-itree.child-num = child-num
            child-num = child-num + 1.
      end.
    end.
      
  end. /* if only child */ 
   
  /* ELSE CREATE PARENT INDEX-TREE NODE FOR MULTIPLE CHILDREN */
  else do:
    create t-itree.
    assign
      t-itree.iid = p-node-iid
      t-itree.parent-iid = p-parent-iid
      t-itree.child-num = child-num
      t-itree.type = pt-wtree.type
      t-itree.inode-wid = pt-wtree.inode-wid
      t-itree.orig-wid = pt-wtree.wid
      t-itree.word-wid = (if pt-wtree.word-index = "*" then -1 else 0)
      t-itree.is-fem = pt-wtree.type = "AND".
    if not can-find( first t-tindex where 
             t-tindex.inode-wid = t-itree.orig-wid )
    then
      pt-wtree.i-state = "+".
    for each child-itree where child-itree.parent-iid = p-node-iid:
      if not child-itree.is-fem or child-itree.index-name = "recid" then
        t-itree.is-fem = no.
      if t-itree.type = "OR" and child-itree.whole-wid > 0 then
        t-itree.whole-wid = child-itree.whole-wid.
    end.
  end.
  
          
end procedure.


/***************************************************************************
*                    PROCEDURE ANALYSE-WHERE-UNDO
* 
* An inode that has already been analysed may need to be undone for some
*  reason.
* If p-whole-wid > 0 then a sibling of this inode has subsequently been 
*  found to be whole-index (either a younger sibling, or an older sibling
*  from a higher generation this inode is being merged into). This inode
*  needs to be completely excluded (its where-tree children will be marked
*  as excluded later in "tidy-where-tree") and any attached index-tree
*  records must be deleted (only one level of deletion is necessary because
*  if more levels exist then it must have a word-index component and
*  therefore can't be whole-index) along with attached index analysis.
* Otherwise, partial undo may be necessary if an AND-node has been replaced
*  by an only-child OR-node. In this case, the AND-node is skipped rather
*  than excluded and only the index analysis needs to be undone.

***************************************************************************/

procedure analyse-where-undo:
  def input param p-inode-wid as int.
  def input param p-whole-wid as int.

  def buffer t-wtree for t-where-tree.
  def buffer t-itree for t-index-tree.
  def buffer t-tindex for t-test-index.
  def buffer t-set for t-test-fem-set.

  if p-whole-wid > 0 then
  do:
    for each t-wtree where t-wtree.inode-wid = p-inode-wid:
      assign
        t-wtree.field-name = ""
        t-wtree.i-state = "X"
        t-wtree.comment = 
          "whole-index [" + string( p-whole-wid, "99" ) + "]".
    end.
  
  end.
   
  for each t-itree where t-itree.inode-wid = p-inode-wid:
    delete t-itree.
  end.

  /* DELETE ALL INODE INDEX ANALYSIS INCLUDING FEM-SET ANALYSIS */
  for each t-tindex where t-tindex.inode-wid = p-inode-wid:
    delete t-tindex.
  end.
  for each t-set where t-set.inode-wid = p-inode-wid:
    delete t-set.
  end.
  
end procedure.


/***************************************************************************
*                    PROCEDURE ANALYSE-INDEXES
* 
* Test indexes are created and analysed for the current inode prior to index
*  selection.
* A test-index will be created for each non-word index of the table.
* A pseudo test-index will be created for RECID if inode has a RECID/ROWID
*  phrase.
* A pseudo test-index will be created for {FEM-set} if AND-inode and there
*  is at least one non-unique FEM test-index (see "analyse-fem-set").
* Each table test-index is analysed for number of equality, range, and sort
*  matches by examining each field of the index in turn.
* Equality and range matches are found in where-tree phrases of the inode.
* An equality match for the field is looked for first. If one isn't found,
*  then look for range matches. There may be up to two range matches for a
*  field. If an equality match isn't found then no more fields are examined.
* The test for sort matching remains alive if the first BY field doesn't
*  match index fields so long as equality match testing is still alive. Once
*  the first BY field matches an index field then each following BY field
*  much match the current index field to keep sort match testing alive.
* If a table index has equality matches for all its fields then it is "FEM"
*  (Full Equality Match) and it will have a fem type of "U" if unique or
*  "N" if non-unique. A RECID test-index will have a fem type of "R", a
*  FEM-set test-index will have a fem type of "S", and a non-FEM test-index
*  will have no fem type (fem = "").
***************************************************************************/

procedure analyse-indexes:
  def input param p-inode-wid as int.
  def input param p-and as log.
  
  def var check-bracket as log.
  def var check-sort as log.
  def var by-pos as int.
  def var num as int.
  
  def buffer t-index for t-index.
  def buffer t-ixfld for t-ixfld.
  def buffer t-wtree for t-where-tree.
  def buffer t-tindex for t-test-index.
 
  def query q-match for t-wtree.
  
  /* ANALYSE EACH NON-WORD INDEX OF CURRENT TABLE */
  for each t-table where t-table.table-name = table-name,
      each t-index where 
           t-index.table-name = table-name and
           t-index.index-type <> "W":

    /* CREATE TEST-INDEX RECORD FOR TABLE INDEX */
    create t-tindex.
    assign
      t-tindex.inode-wid = p-inode-wid
      t-tindex.index-order = t-index.index-order
      t-tindex.index-name = t-index.index-name
      t-tindex.index-type = t-index.index-type
      t-tindex.is-primary = t-index.index-name = t-table.primary-index
      check-bracket = yes
      check-sort = by-fields <> ""
      by-pos = 1.

    /* EXAMINE EACH FIELD OF INDEX IN TURN */
    for each t-ixfld where
             t-ixfld.table-name = table-name and
             t-ixfld.index-name = t-index.index-name:
      
      /* DETERMINE NUMBER OF EQUALITY AND RANGE MATCHES FOR TEST INDEX */
      if check-bracket then
      do:
        num = 0.
        for each t-wtree where
                 t-wtree.inode-wid = p-inode-wid and
                 t-wtree.field-name = t-ixfld.field-name and
                 t-wtree.e-state begins "I".
          num = num + 1.
          if t-wtree.e-state = "IEQ" then
            t-tindex.num-equality = t-tindex.num-equality + 1.
          else
            assign
              t-tindex.num-range =
                t-tindex.num-range +
                (if t-wtree.e-state = "IRB" then 2 else
                 if t-wtree.e-state begins "IR" then 1 else 0)
              check-bracket = no.
        end.
        if num = 0 then
          check-bracket = no.        
      end.
      
      /* DETERMINE NUMBER OF SORT MATCHES FOR TEST INDEX */
      if check-sort then
      
        /* CHECK IF CURRENT BY-FIELD MATCHES CURRENT INDEX FIELD */
        if num-entries( by-fields, " " ) >= by-pos and
           entry( by-pos, by-fields, " " ) = t-ixfld.field-name
        then
          assign
            t-tindex.num-sort = t-ixfld.field-order
            by-pos = by-pos + 1.
        
        /* CHECK-SORT REMAINS ALIVE IF STILL EQUALITY AND FIRST BY-FIELD */
        else 
        if by-pos > 1 or not check-bracket
        then
          check-sort = no.
      
      /* LEAVE IF NOTHING MORE TO CHECK */   
      if not (check-bracket or check-sort) then
        leave.
         
    end. /* for each t-ixfld */

    /* IF STILL CHECK-BRACKET THEN INDEX IS FEM */
    if check-bracket then
      t-tindex.fem = t-tindex.index-type.
       
  end. /* for each t-table, t-index */

  /* CREATE PSEUDO TEST-INDEX FOR RECID IF INODE HAS RECID/ROWID PHRASE */
  if can-find( first t-wtree where
       t-wtree.inode-wid = p-inode-wid and
       t-wtree.field-name = "RECID" ) 
  then do:
    create t-tindex.
    assign
      t-tindex.inode-wid = p-inode-wid
      t-tindex.index-name = "RECID"
      t-tindex.fem = "R".
  end.
  
  /* CREATE FEM-SET PSEUDO TEST-INDEX IF AND-INODE AND NOT V6 QUERY */   
  if p-and and not v6-query then
    run analyse-fem-sets( p-inode-wid ).
        
end procedure.


/***************************************************************************
*                    PROCEDURE ANALYSE-FEM-SETS
* 
* The complete list of non-unique FEM indexes for this inode is created by
*  index order.
* Each combination of this list is analysed to find the "best" FEM-set to
*  create for this inode. 
* The FEM-set is a single "pseudo-index" entity that will compete with all
*  the other indexes for selection in procedure "select-indexes".
* A FEM-set is only created if at least two non-unique FEM indexes.
* All the rules for choosing the best FEM-set are incorporated in a single
*  "FOR EACH t-set" statement which sorts each candidate set so that the
*  first one obeys the following sequence of rules:
* (1) Narrow the selection to those sets which cover the maximum number of
*     distinct fields. If only one set remains, choose this.
* (2) Narrow the selection to those sets which have the minimum number of
*     indexes. If only one set remains, choose this.
* (3) Narrow the selection to those sets which have the greatest sum of
*     index fields (including duplicate fields). If only one set remains,
*     choose this.
* (4) If more than one set remains, choose the one with the lowest
*     combi-num (based on index-order of indexes).
* NOTE:
* - There is no such thing as a FEM-set in the index-tree constructed by the
*  compiler so it is questionable whether it actually exists, but I think
*  this procedure together with rule (6) in "select-indexes" best explains
*  the selection of one or more non-unique-FEM indexes. The set provides the
*  reason why they are chosen, but members are added to the index-tree
*  individually.
* - This procedure should only be run if the inode is an AND-node, not a
*  leaf. A single non-unique FEM index will be selected ahead of a non-FEM 
*  index with the same number of equality matches and range matches no
*  matter how many sort matches they have if AND-inode and not V6 query. If
*  the inode is a leaf then the sort matches rule applies. This is perhaps
*  the best evidence that Progress creates something like a FEM-set for one
*  or more non-unique indexes - but not if leaf inode. Simply adding 
*  "AND YES" to a WHERE expression can affect index selection.
* - There is no need to examine combinations for greater r values after a
*  set with maximum field-list has been found, but a test-set is created 
*  for each combination for report analysis purposes.
***************************************************************************/

procedure analyse-fem-sets:
  def input param p-inode-wid as int.
  
  def var i as int.
  def var j as int.
  def var r as int.
  def var index-list as char.
  def var combi-list as char.
  def var candidate-set as char.
  def var field-list as char.
  def var sum-fields as int.
  def var combi-num as int.
  
  def buffer t-tindex for t-test-index.
  def buffer t-set for t-test-fem-set.
  def buffer t-index for t-index.
  def buffer t-ixfld for t-ixfld.
  
  /* MAKE LIST OF NON-UNIQUE FEM (FULL EQUALITY MATCH) INDEXES */
  for each t-tindex where 
           t-tindex.inode-wid = p-inode-wid and 
           t-tindex.fem = "N"
      by t-tindex.index-order:
    if lookup( t-tindex.index-name, index-list ) = 0 then
      index-list = trim( index-list + "," + t-tindex.index-name, "," ).
  end.
  
  /* IF NO NON-UNIQUE FEM INDEXES THEN NO FEM-SET IS CREATED */
  if index-list = "" then
    return.
  
  /* EXAMINE EACH COMIBATION OF THE COMPLETE LIST IN TURN */  
  do r = 1 to num-entries( index-list ):
    combi-list = combination( index-list, r ).
    do i = 1 to num-entries( combi-list, " " ):
      assign
        candidate-set = entry( i, combi-list, " " )
        field-list = ""
        sum-fields = 0.

      /* CREATE FIELD-LIST AND ADD SUM-FIELDS FOR CANDIDATE-SET */
      do j = 1 to num-entries( candidate-set ):
        for each t-index where
                 t-index.table-name = table-name and
                 t-index.index-name = entry( j, candidate-set ),
            each t-ixfld where
                 t-ixfld.table-name = table-name and
                 t-ixfld.index-name = t-index.index-name:
          sum-fields = sum-fields + 1.               
          if lookup( t-ixfld.field-name, field-list ) = 0 then
            field-list = trim( field-list + "," + t-ixfld.field-name, "," ).
        end.
      end.
  
      /* CREATE A TEST-FEM-SET RECORD FOR EACH CANDIDATE */
      create t-set.
      assign
        combi-num = combi-num + 1
        t-set.combi-num = combi-num
        t-set.inode-wid = p-inode-wid
        t-set.index-list = candidate-set
        t-set.field-list = field-list
        t-set.sum-fields = sum-fields.
      if candidate-set = index-list then
        t-set.key = "max". 

    end.
  end.
  
  /* SORT T-SET TO FIND BEST FEM-SET AND CREATE TEST-INDEX FOR THIS */
  for each t-set where t-set.inode-wid = p-inode-wid
      by num-entries( t-set.field-list ) descending
      by num-entries( t-set.index-list )
      by t-set.sum-fields descending
      by t-set.combi-num:
    create t-tindex.
    assign
      t-tindex.inode-wid = p-inode-wid
      t-tindex.index-name = "~{FEM-set}"
      t-tindex.num-equality = num-entries( t-set.field-list )
      t-tindex.fem-set = t-set.index-list
      t-tindex.fem = "S".
    leave.
  end.
    
end procedure.

  
/***************************************************************************
*                     PROCEDURE SELECT-INDEXES
* 
* Once the where-tree and indexes have been analysed, this procedure
*  selects one or more indexes for an inode.
* An index-tree leaf is created for each CONTAINS phrase of the inode.
* All the rules for selecting a test-index are incorporated in a single
*  "FOR EACH t-tindex" statement. All test-indexes (created in procedure
*  "analyse-test-indexes") will be considered unless a word-index or
*  OR-type index is already attached to the inode in which case all non-FEM
*  indexes are excluded from consideration. These are all sorted so that 
*  the first one will fit the following sequence of rules:
* (1) If USE-INDEX option then select specified index.
* (2) If RECID/ROWID phrase then select RECID pseudo-index.
* (3) If there is a Unique-FEM index then select the last one by index
*     order.
* (4) Narrow selection to those that have the greatest number of Equality
*     Matches. If only one index remains then select this.
* (5) Narrow selection to those that have the greatest number of Range
*     Matches. If only one index remains then select this.
* (6) If the FEM-set is in this selection, then select this.
* (7) Narrow selection to those that have the greatest number of Sort
*     Matches. If only one index remains then select this.
* (8) If the Primary index is in this selection then select this -
*     unless whole-index and at least one Sort Match in which case ignore
*     this rule.
* (9) If more than one index still remains then select the first one by
*     index order (alpha or define - depending on index-by-alpha).
* Only the first test-index is selected. An index-tree leaf is created for
*  this unless the FEM-set is selected in which case an index-tree leaf is
*  created for each member of the FEM-set.
* NOTES:
* - It is important to determine if the inode will be whole-index before
*  selection because of the anomaly in rule (8) above.
* - There is a natural division between rules (3) and (4). The first three
*  rules could easily be tested separately with no need to create a FEM-set
*  if one of these rules succeeds, but I prefer to do it this way for
*  reporting
***************************************************************************/

procedure select-indexes:
  def input param p-inode-wid as int.
  def input param p-parent-iid as int.
  def input param p-and as log.
  def input param p-whole-index as log.
  
  def var whole-index as log.
  def var and-inode as log.
  def var child-num as int.
  def var fem-set as char init ?.
  def var reason as char.
  def var pos as int.
    
  def buffer t-wtree for t-where-tree.
  def buffer t-tindex for t-test-index.
  def buffer t-itree for t-index-tree.
  def buffer parent-itree for t-index-tree.
  
  child-num = next-itree-child( p-parent-iid ).
    
  /* EACH CONTAINS PHRASE GETS ITS OWN INDEX-TREE LEAF */
  for each t-wtree where 
           t-wtree.inode-wid = p-inode-wid and
           t-wtree.e-state = "IW":
    create t-itree.
    assign
      t-itree.child-num = child-num
      child-num = child-num + 1
      iid = iid + 1
      t-itree.iid = iid
      t-itree.parent-iid = p-parent-iid
      t-itree.type = "LEAF"
      t-itree.inode-wid = p-inode-wid
      t-itree.index-name = t-wtree.word-index
      t-itree.is-fem = yes
      t-itree.word-wid = t-wtree.wid
      t-itree.reason = "CONTAINS".
  end.
      
  /* SORT CANDIDATE INDEXES AND CHOOSE FIRST ONE */
  for each t-tindex where
           t-tindex.inode-wid = p-inode-wid and
           not (child-num > 1 and t-tindex.fem = "")
      break
      by t-tindex.index-name <> index-use
      by t-tindex.fem <> "R"
      by - (if t-tindex.fem = "U" then t-tindex.index-order else 0)
      by - t-tindex.num-equality
      by - t-tindex.num-range
      by t-tindex.fem <> "S"
      by - t-tindex.num-sort
      by not (if p-whole-index and t-tindex.num-sort > 0 
              then no else t-tindex.is-primary )
      by t-tindex.index-order:
    
    /* IF FIRST T-TINDEX THEN ASSIGN IS-SELECTED, FEM-SET, AND REASON */
    if fem-set = ? then
    assign
      t-tindex.is-selected = yes
      fem-set = t-tindex.fem-set
      reason = 
        (if last( t-tindex.index-order ) then "Only Option" else
         if t-tindex.index-name = index-use then "USE-INDEX" else
         if t-tindex.index-name = "recid" then "RECID/ROWID" else
         if t-tindex.fem = "U" then "Last Unique FEM" else
         if last-of( - t-tindex.num-equality ) then 
           "Most Equality Matches" 
         else
         if last-of( - t-tindex.num-range ) then 
           "Most Range Matches" else
         if t-tindex.fem = "S" then "Is FEM-Set" else
         if last-of( - t-tindex.num-sort ) then "Most Sort Matches" else
         if t-tindex.is-primary and 
            not( p-whole-index and t-tindex.num-sort > 0 )
         then "Primary Index" 
         else 
         if index-by-alpha then "First Alphabetic" 
         else "Defined First").
   
    /* CREATE INDEX-TREE RECORD */
    pos = lookup( t-tindex.index-name, fem-set ).
    if fem-set = "" or pos > 0 then
    do:
      create t-itree.
      assign
        iid = iid + 1
        t-itree.iid = iid
        t-itree.parent-iid = p-parent-iid
        t-itree.child-num = child-num + (if pos > 0 then pos - 1 else 0)
        t-itree.orig-wid = (if p-and then 0 else p-inode-wid)
        t-itree.inode-wid = p-inode-wid
        t-itree.index-name = t-tindex.index-name
        t-itree.type = "LEAF"
        t-itree.is-fem = (t-tindex.fem <> "")
        t-itree.whole-wid = (if p-whole-index then p-inode-wid else 0)
        t-itree.set-flag = 
          (if pos > 0 then 
             1 + (if pos = 1 then 2 else 0) +
             (if pos = num-entries( fem-set ) then 4 else 0)
           else 0)
        t-itree.reason = reason.
    end.
   
    /* LEAVE AFTER FIRST RECORD UNLESS FEM-SET WAS CHOSEN */
    if fem-set = "" then
      leave.
  
  end.
             
end procedure.  


/***************************************************************************
*                      PROCEDURE TIDY-POST-ANALYSIS
* 
* Any processing that may still need to be done after "analyse-where-tree"
*  can be done here.
* Rather than exclude all the descendents of a where-tree node every time a
*  node is excluded for some reason, just do this once for the whole tree
*  here.
* If the index-tree just consists of a single normal index, then any single
*  phrases in the where expression that are covered by this index will be
*  replaced by the compiler with YES (and if the where expression consists
*  of a single phrase that is covered by the index then the where expression
*  will disappear completely). This potentially transformed where expression
*  becomes the filter for the set of records returned by the index process.
*  This transformation becomes significant if a range phrase like "a > 0" is
*  indexed. Unknown-value sorts higher than 0 in the index so this records
*  with a = ? will be returned by the index process but if "a > 0" is in the
*  filter then these records will subsequently be rejected. If "a > 0" is
*  replaced by YES then this rejection won't occur. This transformation does
*  not happen if the index tree is something other than a single normal
*  index. It's not so straightforward to tell in advance if a single index
*  will chosen, so do this here after the index-tree has been created.
***************************************************************************/

procedure tidy-post-analysis:

  def var state as char.
  
  def buffer t-wtree for t-where-tree.
  def buffer t-tindex for t-test-index.
  def buffer t-ixfld for t-ixfld.
  def buffer t-itree for t-index-tree.
  
  /* TIDY WHERE-TREE */
  run tidy-where-tree(
    input root-wid,
    input no
    ).
  
  /* ANY PHRASES COVERED BY SINGLE NORMAL INDEX ARE REPLACED WITH YES */
  for first t-itree where 
            t-itree.iid = root-iid and
            t-itree.type = "LEAF" and 
            t-itree.index-name <> "RECID" and
            t-itree.word-wid = 0:
    for each t-ixfld where
             t-ixfld.table-name = table-name and
             t-ixfld.index-name = t-itree.index-name:
      state = "".
      for each t-wtree where
               t-wtree.inode-wid = root-wid and      
               t-wtree.field-name = t-ixfld.field-name:
        assign
          t-wtree.yes-replace = yes
          state = t-wtree.e-state.
      end.
      if state <> "IEQ" then
        leave.
    end.
  end.
            
end procedure.


/***************************************************************************
*                      PROCEDURE TIDY-WHERE-TREE
* 
* Recursively tidy the where-tree to make sure that any descendent of an
*  excluded node is also explicity excluded.
***************************************************************************/

procedure tidy-where-tree:
  def input param p-wid as int.
  def input param p-exclude as log.
  
  def buffer t-wtree for t-where-tree.
  def buffer child-wtree for t-where-tree.
 
  for first t-wtree where t-wtree.wid = p-wid:
    
    if p-exclude then
      assign
        t-wtree.field-name = ""
        t-wtree.i-state = "X"
        t-wtree.e-state = "X"
        t-wtree.comment = 
          "excluded by " + (if v6-query then "V6-query" else "parent").
            
    for each child-wtree where child-wtree.parent-wid = t-wtree.wid:
      run tidy-where-tree(
        input child-wtree.wid,
        input t-wtree.i-state = "X" or 
              v6-query and not (p-wid = root-wid and t-wtree.type = "AND")
        ).
    end.
    
  end.
  
end procedure.


/***************************************************************************
*                      PROCEDURE REPORT-QIDX-INFO
*
* Report interesting index information with regards to the query, including
*  an analysis of the where-tree, the index-tree, the post-index filter, and
*  index analysis.
***************************************************************************/

procedure report-qidx-info:

  def var qidx-info as char.
  def var where-syntax as char.
  def var index-tree as char.
  def var where-tree as char.
  def var index-analysis as char.
  def var all-analysis as char.
  
  def buffer t-tindex for t-test-index.
  
  form 
    qidx-info view-as editor size 78 by 24 no-word-wrap no-label font 2
    with frame f-report three-d
         view-as dialog-box title "Query Index Information Report".
 
  assign
    qidx-info:width = font-table:get-text-width( fill( "0", 78), 2 ) 
    frame f-report:width = 
      qidx-info:width + 
      frame f-report:border-left + frame f-report:border-right.
      
  run report-where-tree(
    input root-wid,
    input yes,
    input "",
    input 1,
    output where-tree
    ).

  run report-index-tree( 
    input root-iid, 
    input 1, 
    output index-tree
    ).

  run report-where-tree(
    input root-wid,
    input no,
    input "YES",
    input 1,
    output where-syntax
    ).
    
  for each t-tindex break by t-tindex.inode-wid:
    if first-of( t-tindex.inode-wid ) then
    do:
      run report-index-analysis(
        input t-tindex.inode-wid,
        output index-analysis
        ).
      all-analysis = all-analysis + index-analysis.
    end.
  end.
  
  assign
    qidx-info:read-only = yes
    qidx-info =
      "Where Tree Analysis~n" +
      "===================~n~n" +
      where-tree + "~n" +
      "Index Tree~n" +
      "==========~n~n" +
      index-tree + "~n" +
      "Post-Index Filter~n" +
      "=================~n~n" +
      where-syntax + "~n~n" +
      "Index Analysis~n" +
      "==============~n" +
      all-analysis
      .
    
  update qidx-info with frame f-report.
    
end procedure.


/***************************************************************************
*                      PROCEDURE REPORT-WHERE-TREE
* 
* Recursively report the where expression, either in tree format (if p-tree)
*  or syntax format (if not p-tree).
* The tree format reports the wid of each node (for reference with comments,
*  index-tree, and index analysis, etc), the i-state, and an analytical
*  comment that hopefully helps to explain why the index-tree was
*  constructed the way it was. See comments for "define temp-table
*  t-where-tree" for a description of possible i-state values.
* The syntax format reports the where-tree data as syntactic text similar to
*  how the where-expr was originally entered. With the default mode (p-mode
*  is blank), each single phrase of the expression is reported as a separate
*  line. 
* If p-mode = "convert" it is reported as a single line with clean
*  whitespace and converted where-tree data is then deleted. 
* If p-mode = "yes" and t-wtree.yes-replace then the expr is replaced with
*  YES followed by the expression in comments (or "EMPTY" comment if it is
*  the root node). This is intended to replicate the post-index query filter
*  in the compiled code. This is significant if a range expression is
*  replaced by YES and unknown values are involved.
***************************************************************************/

procedure report-where-tree:
  def input param p-wid as int.
  def input param p-tree as log.
  def input param p-mode as char.
  def input param p-level as int.
  def output param p-text as char.
  
  def var child-text as char.
  def var yes-mode as log.
  def var comment as char.

  def buffer t-wtree for t-where-tree.
  def buffer child-wtree for t-where-tree.
  
  find t-wtree where t-wtree.wid = p-wid no-error.
  if not available t-wtree then
    return.
  
  if p-tree then
    p-text = 
      p-text +
      "[" + string( t-wtree.wid, "99" ) + "]" + fill( "+-", p-level ) +
      t-wtree.type + " [" + t-wtree.i-state + "] " +
      t-wtree.expr + 
      (if t-wtree.comment = "" then ""
       else "   /" + "* " + t-wtree.comment + " *" + "/") +
      "~n".

  /* REPORT GROUP NODE */
  if t-wtree.type <> "LEAF" then
  do:
    
    /* REPORT EACH CHILD RECURSIVELY */
    for each child-wtree where
             child-wtree.parent-wid = p-wid:
      run report-where-tree(
        input child-wtree.wid,
        input p-tree,
        input p-mode,
        input p-level + 1,
        output child-text
        ).
      p-text = 
        p-text + 
        (if not p-tree and p-text <> "" 
         then " " + t-wtree.type + "~n"
         else "") +
        child-text.
    end.
    
    /* ADD BRACKETS AROUND INNER OR-PHRASE FOR SYNTAX MODE */
    if not p-tree and t-wtree.type = "OR" and p-level > 1 
    then
      p-text = "(" + p-text + ")".
      
  end.
  
  /* REPORT LEAF DETAILS FOR SYNTAX MODE */
  else
  if not p-tree then 
    assign
      yes-mode = p-mode = "YES" and t-wtree.yes-replace
      p-text =
        p-text +
        (if yes-mode then 
           if t-wtree.wid = root-wid then "/" + "* EMPTY was: "
           else "YES /" + "* was: "
         else "") +
        t-wtree.expr + 
        (if yes-mode then " *" + "/" else "").
                
  /* IF CONVERT MODE THEN CLEAN WHITESPACE AND DELETE WTREE */ 
  if p-mode = "CONVERT" then
  do:
    p-text = clean-whitespace( p-text ).
    delete t-wtree.
  end.
  
end procedure.

    
/***************************************************************************
*                      PROCEDURE REPORT-INDEX-TREE
* 
* Recursively report the index-tree. Also report field details for each
*  selected index and the reason why it was selected.
* A FEM-set is never part of the index tree but it is included in the report
*  because the reason why its members were selected is connected to this.
* The [fem], [word], and [whole] flags are intended to replicate flags in
*  the compiled index-tree (which show as "FGE", "QBW" and "WORD-INDEX" in
*  -zqilv output). The [word] flag is attached to any node that is a
*  word-index or has a word-index descendant. The [whole] flag is attached
*  to any individual whole-index and any OR-node with a whole-index child.
*  The [fem] flag is attached to any FEM index, word-index, RECID pseudo
*  index, and AND-node whose children are all FEM-indexes or word-indexes 
*  (but not RECID for some reason, even though RECID has [fem] flag). 
* Note that the [whole] flag refers to the prog-whole version of whole-index
*  rather than true whole-index (see comments in "analyse-where-tree"). True
*  whole-index can be seen in this report when an index has "---" as a 
*  bracket expression for all its fields.
* The compiled index bracket is somewhat different in structure from the
*  field details reported here but the information is essentially the same.
***************************************************************************/

procedure report-index-tree:
  def input param p-iid as int.
  def input param p-level as int.
  def output param p-text as char.
  
  def var child-text as char.
  def var inode as char.
  def var num as int.
    
  def buffer t-itree for t-index-tree.
  def buffer child-itree for t-index-tree.
  def buffer t-wtree for t-where-tree.
  
  find t-itree where t-itree.iid = p-iid no-error.
  if not available t-itree then
    return.

  /* REPORT INDEX-TREE NODE HEADER DETAILS */
  assign
    inode = "[" + string( t-itree.inode-wid, "99" ) + "]"
    p-text =  
      (if get-bits( t-itree.set-flag, 2, 1 ) = 1 then
         inode + fill( "  ", p-level ) + "~{FEM-set selected because: " +
         t-itree.reason + "~n"
       else "") +
      inode + fill( "+-", p-level ) + 
      (if t-itree.type = "LEAF" then
         (if t-itree.index-name <> "RECID" then "INDEX " else "") +
         t-itree.index-name
       else t-itree.type) +
      (if t-itree.is-fem then " [fem]" else "") +
      (if t-itree.word-wid <> 0 then " [word]" else "") +
      (if t-itree.whole-wid > 0 then " [whole]" else "") +
      (if t-itree.type = "LEAF" then
         if t-itree.set-flag = 0 
         then " selected because: " + t-itree.reason
         else " is member of FEM-set"
       else "") +  
      "~n".
    
  /* REPORT RECID DETAILS */         
  if t-itree.index-name = "RECID" then
    for each t-wtree where
             t-wtree.inode-wid = t-itree.inode-wid and
             t-wtree.field-name = "RECID" and
             t-wtree.e-state = "IID":
      p-text = 
        p-text + inode + fill( "  ", p-level + 1 ) + t-wtree.expr + "~n".
    end.
    
  /* REPORT INDEX FIELD DETAILS */
  for each t-ixfld where
           t-ixfld.table-name = table-name and
           t-ixfld.index-name = t-itree.index-name:
      p-text = p-text + inode + fill( "  ", p-level + 1 ) + "FIELD ". 
    
    /* USE WORD-WID IF WORD-INDEX */
    if t-itree.word-wid > 0 then
      for each t-wtree where t-wtree.wid = t-itree.word-wid:
        p-text = p-text + t-wtree.expr + "~n".
      end.
      
    /* ELSE NORMAL INDEX */
    else do:
      num = 0.
      for each t-wtree where
               t-wtree.inode-wid = t-itree.inode-wid and
               t-wtree.field-name = t-ixfld.field-name and
               t-wtree.e-state begins "I":
        assign
          num = num + 1
          p-text = 
            p-text + 
            (if num > 1 then inode + fill( "  ", p-level + 4 ) else "") + 
             t-wtree.expr + "~n".
      end.
      if num = 0 then
        p-text = p-text + t-ixfld.field-name + " ---~n". 
    end.
    
  end. /* for each t-ixfld */

  /* CLOSE BRACKET IF LAST MEMBER OF FEM-SET */
  if t-itree.set-flag > 4 then  
    substring( p-text, length( p-text ), 0 ) = "}".
  
  /* RECURSIVELY REPORT CHILDREN */  
  for each child-itree where child-itree.parent-iid = p-iid:
    run report-index-tree(
      input child-itree.iid,
      input p-level + 1,
      output child-text
      ).
    p-text = p-text + child-text.
  end.

  
end procedure.


/***************************************************************************
*                    PROCEDURE REPORT-INDEX-ANALYSIS
* 
* Report an analysis of test-indexes and test-fem-sets for a particular
*  inode.
* Non-excluded non-word indexes are listed in order with respect to the
*  following columns (all descending order except "O#"):
*  UI: USE-INDEX this index? (Y/N)
*  ID: RECID pseudo-index (for RECID/ROWID)? (Y/N)
*  FU: Index-Order number if Unique FEM index
*  EM: number of Equality Matches
*  RM: number of Range Matches
*  FS: FEM-set pseudo-index? (Y/N)
*  SM: number of Sort Matches.
*  Pr: Primary Index (and not whole-index anomaly)? (Y/N)
*  O#: Index-Order number
* The index which sorts highest will be selected and marked with an "*". The
*  reason it was selected is determined by the left-most column whose value
*  is different from that of the index that sorts second highest.
* Each index that has been excluded from selection consideration, because
*  it is non-FEM and at least one word-index or OR-child has already been
*  selected for that inode, will be listed with a comment without column
*  values.
* Each word-index is listed with a count of its index-tree leaves for this
*  inode. A count greater than zero means it is selected.
* A count of OR-childs is also provided if at least one exists.
* If a FEM-set was created then an analysis of test FEM-sets is provided.
*  The complete list of Non-Unique FEM indexes and all the distinct fields
*  they cover is reported in the header and then combinations are listed in
*  order with respect to following columns:
*  NF: Number of distinct fields covered by the set (descending)
*  NI: Number of indexes in the set (ascending)
*  SF: Sum total of fields covered by the set (descending)
*  FEM-set: based on index-order of indexes in the set (ascending).
* There could potentially be thousands of test FEM-sets so no more than the
*  top 10 are reported to avoid maxing out the 32K limit for p-text.
***************************************************************************/

procedure report-index-analysis:
  def input param p-inode-wid as int.
  def output param p-text as char.
  
  def var whole as log.
  def var word as log.
  def var fem-set as char.
  def var num as int.
  
  def buffer t-index for t-index.
  def buffer t-wtree for t-where-tree.
  def buffer t-itree for t-index-tree.
  def buffer t-tindex for t-test-index.
  def buffer t-set for t-test-fem-set.
    
  /* DETERMINE IF THIS INODE HAS FEM-SET OR IS WHOLE-INDEX */
  find 
    t-tindex where 
    t-tindex.inode-wid = p-inode-wid and 
    t-tindex.fem = "S"
    no-error.
  assign
    fem-set = (if available t-tindex then t-tindex.fem-set else "")
    word = 
      can-find( first t-itree where  
        t-itree.inode-wid = p-inode-wid and 
        t-itree.word-wid <> 0 )
    whole = 
      can-find( t-itree where 
        t-itree.orig-wid = p-inode-wid and 
        t-itree.whole-wid > 0 ).

  /* INDEX ANALYSIS HEADER */
  assign
    p-text = 
      "~n" + "Index Analysis for [" +
      string( p-inode-wid, "99" ) + "]:~n" +
      "  Index name                       UI ID FU EM RM FS SM Pr O#~n" +
      "  -------------------------------- -- -- -- -- -- -- -- -- --~n".
  
  /* REPORT ON ALL NON-WORD INDEXES (INCLUDING PSEUDO-INDEXES) */
  for each t-tindex where t-tindex.inode-wid = p-inode-wid
      by t-tindex.fem = "" and word
      by t-tindex.index-name <> index-use
      by t-tindex.fem <> "R"
      by - (if t-tindex.fem = "U" then t-tindex.index-order else 0)
      by - t-tindex.num-equality
      by - t-tindex.num-range
      by t-tindex.fem <> "S"
      by - t-tindex.num-sort
      by not (if whole and t-tindex.num-sort > 0 
          then no else t-tindex.is-primary )
      by t-tindex.index-order:
    assign
      p-text =
        p-text + (if t-tindex.is-selected then "* " else "  ") +
        string( t-tindex.index-name, "x(33)" ) + 
        (if t-tindex.fem = "" and word then
           "(Non-FEM excluded by Word/OR)"
         else
           " " +
           string( t-tindex.index-name = index-use, "Y/N" ) + "  " +
           string( t-tindex.fem = "R", "Y/N" ) + " " +
           string( if t-tindex.fem = "U" then t-tindex.index-order else 0, 
                   ">9" ) + " " +
           string( t-tindex.num-equality, ">9" ) + " " +
           string( t-tindex.num-range, ">9" ) + "  " +
           string( t-tindex.fem = "S", "Y/N" ) + " " +
           string( t-tindex.num-sort, ">9" ) + "  " +
           (if whole and t-tindex.num-sort > 0 then "-" 
            else string( t-tindex.is-primary, "Y/N" )) + " " +
           string( t-tindex.index-order, ">9" )) + 
        "~n".
  
  end. /* for each t-tindex */
    
  /* REPORT ON WORD-INDEXES IN ORDER */
  for each t-index where
           t-index.table-name = table-name and
           t-index.index-type = "W":
    num = 0.
    for each t-wtree where
             t-wtree.inode-wid = p-inode-wid and
             t-wtree.word-index = t-index.index-name:
      assign
        word = yes
        num = num + 1.
    end.
    p-text = 
      p-text + (if num = 0 then "  " else "* ") +
      string( t-index.index-name, "x(33)" ) +
      "(Word-Index count = " + string( num ) + ")~n".
  end.

  /* REPORT ON OR-CHILDREN IF ANY */
  num = 0.
  for each t-itree where
           t-itree.inode-wid = p-inode-wid and
           t-itree.type = "OR":
    num = num + 1.
  end.
  if num > 0 then
    p-text = 
      p-text + "* OR" + fill( " ", 31 ) + "(OR-child count = " +
      string( num ) + ")~n".
    
  /* REPORT ON TEST-FEM-SET (IF THERE IS ONE) */
  if fem-set <> "" then
  do:
  
    /* SHOW ALL NON-UNIQUE FEM INDEXES AND ALL FIELDS THEY COVER */
    for each t-set where
             t-set.inode-wid = p-inode-wid and
             t-set.key contains "max":
      assign
        num = 0
        p-text =
          p-text + "~n" +
          "FEM-Set Analysis for [" +
          string( p-inode-wid, "99" ) + "]:~n" + 
          "   Non-Unique FEM Indexes: " + t-set.index-list + "~n" +
          "    Equality Match Fields: " + t-set.field-list + "~n" +
          "  NF NI SF FEM-Set (leading candidates)~n" + 
          "  -- -- -- ----------------------------~n".
    end.
      
    /* LIST LEADING FEM-SET CANDIDATES IN ORDER */  
    for each t-set where t-set.inode-wid = p-inode-wid
        by num-entries( t-set.field-list ) descending
        by num-entries( t-set.index-list )
        by t-set.sum-fields descending
        by t-set.index-list:
       
      assign
        num = num + 1  
        p-text = 
          p-text +
          (if t-set.index-list = fem-set then "* " else "  ") +
          string( num-entries( t-set.field-list ), ">9" ) + " " +
          string( num-entries( t-set.index-list ), ">9" ) + " " +
          string( t-set.sum-fields, ">9" ) + " " +
          t-set.index-list + "~n".

      /* LEAVE AFTER TEN CANDIDATES SO IT DOESN'T BECOMING TOO LONG */
      if num > 9 then
        leave.
      
    end. /* for each t-set */
    
  end. /* report on fem-set */
  
end procedure.


/***************************************************************************
*                    PROCEDURE REPORT-TABLE-DEF
* 
* Report the definition of a table in tidy temp-table-like syntactic form.
***************************************************************************/

procedure report-table-schema:
  def output param p-text as char.
  
  for each t-table where t-table.table-name = table-name:
    for each t-field where t-field.table-name = table-name:
      p-text = 
        p-text + "FIELD " + t-field.field-name + 
        " AS " + t-field.data-type + 
        (if t-field.extent > 0 
         then " EXTENT" + string( t-field.extent ) else "") + "~n".
    end.
    for each t-index where t-index.table-name = table-name
        by t-index.index-order:
      p-text = 
        p-text + "INDEX " + t-index.index-name +
        (if t-index.index-name = t-table.primary-index 
         then " IS PRIMARY" else "") +
        (if t-index.index-type = "U" then " IS UNIQUE" else "") +
        (if t-index.index-type = "W" then " IS WORD-INDEX" else "") + 
        (if t-index.num-fields > 1 then "~n" else "").
      for each t-ixfld where
               t-ixfld.table-name = table-name and 
               t-ixfld.index-name = t-index.index-name:
        p-text = 
          p-text + "  " + t-ixfld.field-name + " " + 
          string( t-ixfld.is-ascending, "ASCENDING/DESCENDING" ) + "~n".
      end.
    end.
  end.

end procedure.

  
/**************************************************************************/

